#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define N 50
#define NG 50
#define pi 3.14159265

#define EPS_MIN 1e-7
#define EPS_ANG 1e-9
#define EPS_NEWT 1e-11
#define GET_X_gamma(xs) ((35.0+xs*(92.0 + xs*17.0))/(105.0+xs*(24.0+15.0*xs)))

/************************************************************

  get_gamma(X)

************************************************************/
double get_gamma(X)
double X;
{
  double f, fp, g, del, eps, y;
  
  g = X;
  eps = 1.0;
  if (X < EPS_MIN) {return(3.0*X);}
  if (fabs(X - 1.0) < EPS_MIN) {return(1.0);}
  while(eps > EPS_NEWT){
    y = log((1+g)/(1-g));
    f = X-1/g+2.0/y;
    fp = -1/(g*g)-4.0/((1-g*g)*y*y);
    del = -f/fp;
    eps = (fabs(del/g));
    while ((g + del) >= 1.0) {
      del /= 2.0;
    }
    g = g + del;      
  }
  return(g);
}
/************************************************************

  get_X_gamma(Xs)

  called with Xs = X*X

************************************************************/
double get_X_gamma(Xs)
double Xs;
{
  static int First_Time = 1;
  static double X_g[NG+1], X_gprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double xs, x, x_g, g, gprime;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      xs = (double)i/dN;
      x = sqrt(xs);
      g = get_gamma(x);
      if (x < EPS_MIN){
	gprime = 3.0 ; 
	X_g[i] = .33333333333333333333333;
	X_gprime[i] = 0.8;
      } else {
	X_g[i] = x_g = x/g;
	gprime = (x < (1.0-EPS_MIN))?
	  ((1-g*g)/(1 + x*x - 2*x_g)):0.0;
	X_gprime[i] = .5*(1-x_g*gprime)/(g*x);
      }
      X_gprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(X_gprime[i]+X_gprime[j])+.25*(X_g[i]-X_g[j]);
      b[i] = .25*(X_gprime[j]-X_gprime[i]);
      c[i] = -.25*(X_gprime[i]+X_gprime[j])-.75*(X_g[i]-X_g[j]);
      d[i] = .5*(X_g[i]+X_g[j])+.25*(X_gprime[i]-X_gprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*Xs);
  if (i>=NG) {
    g = .5*(Xs+1);
    return(g);
  }
  j = i+1;
  x = (Xs+Xs)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}
/************************************************************

  get_g(X)

************************************************************/
double get_g(X)
double X;
{
  static int First_Time = 1;
  static double gamma[NG+1], gprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double x, g, di;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      x = (double)i/dN;
      gamma[i] = g = get_gamma(x);
      gprime[i] = (g < EPS_MIN)?
	3.0:
	(g < (1.0-EPS_MIN))?
	(g*(1-g*g)/(g-2*x+g*x)):
	0.0;
      gprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(gprime[i]+gprime[j])+.25*(gamma[i]-gamma[j]);
      b[i] = .25*(gprime[j]-gprime[i]);
      c[i] = -.25*(gprime[i]+gprime[j])-.75*(gamma[i]-gamma[j]);
      d[i] = .5*(gamma[i]+gamma[j])+.25*(gprime[i]-gprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*X);
  if (i >= NG){
    g = 1.0;
    return(g);
  }
  j = i+1;
  x = (X+X)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}
/************************************************************

  get_Pterm(Xs)

  called with Xs = X*X

************************************************************/
double get_Pterm(Xs)
double Xs;
{
  static int First_Time = 1;
  static double Pterm[NG+1], Ptermprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double x, xs, x_g, g, gprime;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      xs = (double)i/dN;
      x = sqrt(xs);
      g = get_gamma(x);
      if (x < EPS_MIN){
	gprime = 3.0 ; 
	Pterm[i] = .33333333333333333333333;
	Ptermprime[i] = -0.4;
      } else {
	x_g = x/g;
	gprime = (x < (1.0-EPS_MIN))?
	  ((1-g*g)/(1 + x*x -2*x_g)):0.0;
	Pterm[i] = .5*(1-x_g);
	Ptermprime[i] = .25*(x_g*gprime-1)/(x*g);
      }
      Ptermprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(Ptermprime[i]+Ptermprime[j])+.25*(Pterm[i]-Pterm[j]);
      b[i] = .25*(Ptermprime[j]-Ptermprime[i]);
      c[i] = -.25*(Ptermprime[i]+Ptermprime[j])-.75*(Pterm[i]-Pterm[j]);
      d[i] = .5*(Pterm[i]+Pterm[j])+.25*(Ptermprime[i]-Ptermprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*Xs);
  if (i>=NG) {
    g = .25*(1-Xs);
    return(g);
  }
  j = i+1;
  x = (Xs+Xs)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}
/************************************************************

  get_Vterm(Xs)

  called with Xs = X*X

************************************************************/
double get_Vterm(Xs)
double Xs;
{
  static int First_Time = 1;
  static double Vterm[NG+1], Vtermprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double x, xs, x_g, g, gprime;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      xs = (double)i/dN;
      x = sqrt(xs);
      g = get_gamma(x);
      if (x < EPS_MIN){
	gprime = 3.0 ; 
	Vterm[i] = 6.0/5.0;
	Vtermprime[i] = -18.0/(165.0);
      } else {
	x_g = x/g;
	gprime = (x < (1.0-EPS_MIN))?
	  ((1-g*g)/(1 + x*x - 2*x_g)): 0.0;
	Vterm[i] = (1.5/g-.5/x)/x;
	Vtermprime[i] = .5*(1/(xs*xs)-1.5*(1/x+gprime/g)/(g*xs));
      }
      Vtermprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(Vtermprime[i]+Vtermprime[j])+.25*(Vterm[i]-Vterm[j]);
      b[i] = .25*(Vtermprime[j]-Vtermprime[i]);
      c[i] = -.25*(Vtermprime[i]+Vtermprime[j])-.75*(Vterm[i]-Vterm[j]);
      d[i] = .5*(Vterm[i]+Vterm[j])+.25*(Vtermprime[i]-Vtermprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*Xs);
  if (i>=NG) {
    g = 1.25-.25*Xs;
    return(g);
  }
  j = i+1;
  x = (Xs+Xs)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}
/************************************************************

  main()

 ************************************************************/
main()
{
  double g, gr, gz, dN, z, s, y;
  double zn, zGr, zGz;
  double theta, X, Xs, Xog, t1, t2, t3;
  int i, j, limit;
  enum g_region {normal, r_axis, z_axis, edge_normal, edge_zaxis, center};
  typedef enum g_region region;
  region current_point;


  dN = (double)N; 

  g = get_g(0.0);

  for(i=0; i<=310; i++){
    Xs = (double)i/300;
    s = get_Vterm(Xs);
    y = (35+92*Xs+17*Xs*Xs)/(105+24*Xs+15*Xs*Xs);
    y = (i>0)?(.5*(3*y-1)/Xs):(6.0/5.0);
    /*    y = .5*(1-y);*/
    fprintf(stdout,"%8.5e %8.5e %8.5e\n",Xs, s, y);
  }

#if 0
  for(i=0; i<=N; i++){
    g = (double)i/dN;
    if (i == 0) {
      X = 0.0;
    } else {
      X = (g < (1.0-EPS_MIN))?(1.0/g - 2.0/log((1+g)/(1-g))):1.0;
    }

    fprintf(stderr,"Input X: ");
    fscanf(stdin,"%lf",&X);
    Xs = X*X;
    g = get_g(X);
    /*    g = (i == 0)?0.0:X/GET_X_gamma(Xs);*/
    /*    g = get_gamma(X);*/
    g = (g > 1.0)?1.0:g;

    for(j=-N; j<=N; j++){
      theta = (double)(j)*pi/dN;
      gr = g*((cos(theta)));
      gz = g*sin(theta);
#if 0    
      fprintf(stderr,"gr: ");
      fscanf(stdin,"%lf",&gr);
      fprintf(stderr,"gz: ");
      fscanf(stdin,"%lf",&gz);
#endif
      s = sqrt(1-gz*gz);

      current_point = normal;
      if (g >= (1.0-EPS_MIN)) {
	if (gr*gr > EPS_ANG){
	  current_point = edge_normal;
	} else {
	  current_point = edge_zaxis;
	}
      } else if (g <= EPS_MIN) {
	current_point = center;
      } else if (gz*gz <= EPS_ANG) {
	current_point = r_axis;
      } else if (gr*gr <= EPS_ANG) {
	current_point = z_axis;
      }
      switch (current_point){
      case normal:
	t1 = log(((g-gr)*(g*s+gr))/((g+gr)*(g*s-gr)));
	zn = ((1-g*X)/(g*g))*
	  ((1-s)+
	  .5*(gr/g)*t1
	   );
	zGr = ((1-g*X)/(g*X))*
	  (1.5*(1-s)/(g*g)-0.5 +
	   (0.5*gz*gz+1.5*gr*gr/(g*g)-0.5)*t1
	   /(2*g*gr)
	   );
	zGz = ((1-g*X)/(g*X))*
	  ((g*g-1.5*gr*gr)/(g*g*(1+s)) - 0.5 + 
	   .25*(gr/g)*(3.0/(g*g)-1)*t1
	   );
	  break;
      case edge_normal:
	y = fabs(gr/g);
	zn = ((1-g*X)*(1-s+y*(t1=log((s+y)/((1+y)*(1+g))))) + y*g)/(g*g);
	t2 = (gz*gz+3*y*y-1)/(g*g*y);
	zGr = ((1-g*X)*(3*(1-s)/(g*g)-1+t2*t1))/(2*g*X)+t2/(2*X);
	t3 = y*(3/(g*g)-1);
	zGz = ((1-g*X)*((2-3*y*y)/(1+s)-1+t3*t1)/(2*g*X)) + t3/(2*X);
	break;
      case edge_zaxis:
	zn = (1-s)*(1-g*X)/(g*g);
	zGr = .5*((1-g*X)/(g*X))*((1-s)*(3-s)/(g*g)-1.0);
	zGz = .5*((1-g*X)/(g*X))*((1-s)/(1+s));
	break;
      case center:
	zn = 0.5;
	zGr = 0.75;
	zGz = 0.375;
	break;
      case r_axis:
	zn = (log(1-g*g)/(log(1-g)-log(1+g)))/g;
	zGr = -.5*((1-g*X)/(g*X))*(g*g + log(1-g*g))/(g*g);
	zGz = -.5*((1-g*X)/(g*X))*(1.5+.5*(3/(g*g)-1.0)*log(1-g*g));
	break;
      case z_axis:
	zn = (1-s)*(1-g*X)/(g*g);
	zGr = .5*((1-g*X)/(g*X))*((1-s)*(3*s+(g*g-1))/(g*g*s)-1.0);
	zGz = .5*((1-g*X)/(g*X))*((1-s)/(1+s));
	break;
      }
      fprintf(stdout,"%8.5e %8.5e %8.5e %8.5e %8.5e\n",theta, zn, zGr, zGz, X);
    }
    fprintf(stdout,"\n");
  }
#endif




}

