/************************************************************

  collisions.c

  Various routines for acquiring collision data and 
  arranging it into various arrays. 
  Also has routine for applying inelastic collision term.
  Part of "poscol" simulation.

************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <cpml.h>
#include "basics.h"
#undef pi
#include "MD_FCT_defs.h"
#include "MD_FCT.h"
#include "MD_FCT_utils.h"
#include "poscol.h"
#include "sources.h"
#include "special.h"

#define NG 50
#define EPS_MIN 1e-7
#define EPS_ANG 1e-7
#define EPS_NEWT 1e-11

typedef struct inelastic_event{
  MD_FCT_species_ptr target;
  MD_FCT_species_ptr product;
  double multiplicity;
  dblptr frequency;
  unsigned_int_ptr jmin;
  dblptr lambda_min;
  unsigned_int_ptr jmax;
  dblptr lambda_max;
  double loss_rate;
  double incr;
  double decr;
} inelastic_event;

typedef inelastic_event *inelastic_event_ptr;

static bool de_excitation(double, axis_ptr, unsigned int, inelastic_event);
static dblptr inelastic = NULL, wall_absorption_rate = NULL, elastic = NULL;
static inelastic_event_ptr excitation_events = NULL, ionization_events = NULL;
dblptr SBC = NULL;

static double get_g(double);
static double get_gamma(double);
double get_Pterm(double);
double get_Vterm(double);
double get_X_Gamma(double);
double do_wall_absorption(dblptr, dblptr, dblptr, double);
/************************************************************

  get_X_gamma(Xs)

  called with Xs = X*X

************************************************************/
double get_X_gamma(Xs)
double Xs;
{
  static int First_Time = 1;
  static double X_g[NG+1], X_gprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double xs, x, x_g, g, gprime;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      xs = (double)i/dN;
      x = sqrt(xs);
      g = get_gamma(x);
      if (x < EPS_MIN){
	gprime = 3.0 ; 
	X_g[i] = .33333333333333333333333;
	X_gprime[i] = 0.8;
      } else {
	X_g[i] = x_g = x/g;
	gprime = (x < (1.0-EPS_MIN))?
	  ((1-g*g)/(1 + x*x - 2*x_g)):0.0;
	X_gprime[i] = .5*(1-x_g*gprime)/(g*x);
      }
      X_gprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(X_gprime[i]+X_gprime[j])+.25*(X_g[i]-X_g[j]);
      b[i] = .25*(X_gprime[j]-X_gprime[i]);
      c[i] = -.25*(X_gprime[i]+X_gprime[j])-.75*(X_g[i]-X_g[j]);
      d[i] = .5*(X_g[i]+X_g[j])+.25*(X_gprime[i]-X_gprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*Xs);
  if (i>=NG) {
    g = .5*(Xs+1);
    return(g);
  }
  j = i+1;
  x = (Xs+Xs)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}
/************************************************************

  get_Pterm(Xs)

  called with Xs = X*X

  Pterm = (1/2)(1-X/gamma)

************************************************************/
double get_Pterm(Xs)
double Xs;
{
  static int First_Time = 1;
  static double Pterm[NG+1], Ptermprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double x, xs, x_g, g, gprime;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      xs = (double)i/dN;
      x = sqrt(xs);
      g = get_gamma(x);
      if (x < EPS_MIN){
	gprime = 3.0 ; 
	Pterm[i] = .33333333333333333333333;
	Ptermprime[i] = -0.4;
      } else {
	x_g = x/g;
	gprime = (x < (1.0-EPS_MIN))?
	  ((1-g*g)/(1 + x*x -2*x_g)):0.0;
	Pterm[i] = .5*(1-x_g);
	Ptermprime[i] = .25*(x_g*gprime-1)/(x*g);
      }
      Ptermprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(Ptermprime[i]+Ptermprime[j])+.25*(Pterm[i]-Pterm[j]);
      b[i] = .25*(Ptermprime[j]-Ptermprime[i]);
      c[i] = -.25*(Ptermprime[i]+Ptermprime[j])-.75*(Pterm[i]-Pterm[j]);
      d[i] = .5*(Pterm[i]+Pterm[j])+.25*(Ptermprime[i]-Ptermprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*Xs);
  if (i>=NG) {
    /*    g = .25*(1-Xs);*/
    g = 0.0;
    return(g);
  }
  j = i+1;
  x = (Xs+Xs)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}
/************************************************************

  get_Vterm(Xs)

  called with Xs = X*X

  Vterm = (3(X/gamma)-1)/(2X*X)

************************************************************/
double get_Vterm(Xs)
double Xs;
{
  static int First_Time = 1;
  static double Vterm[NG+1], Vtermprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double x, xs, x_g, g, gprime;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      xs = (double)i/dN;
      x = sqrt(xs);
      g = get_gamma(x);
      if (x < EPS_MIN){
	gprime = 3.0 ; 
	Vterm[i] = 6.0/5.0;
	Vtermprime[i] = -18.0/(165.0);
      } else {
	x_g = x/g;
	gprime = (x < (1.0-EPS_MIN))?
	  ((1-g*g)/(1 + x*x - 2*x_g)): 0.0;
	Vterm[i] = (1.5/g-.5/x)/x;
	Vtermprime[i] = .5*(1/(xs*xs)-1.5*(1/x+gprime/g)/(g*xs));
      }
      Vtermprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(Vtermprime[i]+Vtermprime[j])+.25*(Vterm[i]-Vterm[j]);
      b[i] = .25*(Vtermprime[j]-Vtermprime[i]);
      c[i] = -.25*(Vtermprime[i]+Vtermprime[j])-.75*(Vterm[i]-Vterm[j]);
      d[i] = .5*(Vterm[i]+Vterm[j])+.25*(Vtermprime[i]-Vtermprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*Xs);
  if (i>=NG) {
    /*    g = 1.25-.25*Xs;*/
    g = 1.0;
    return(g);
  }
  j = i+1;
  x = (Xs+Xs)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}
/************************************************************

  get_gamma(X)

************************************************************/
static double get_gamma(X)
double X;
{
  double f, fp, g, del, eps, y;
  
  g = X;
  eps = 1.0;
  if (X < EPS_MIN) {return(3.0*X);}
  if (fabs(X - 1.0) < EPS_MIN) {return(1.0);}
  while(eps > EPS_NEWT){
    y = log((1+g)/(1-g));
    f = X-1/g+2.0/y;
    fp = -1/(g*g)-4.0/((1-g*g)*y*y);
    del = -f/fp;
    eps = (fabs(del/g));
    while ((g + del) >= 1.0) {
      del /= 2.0;
    }
    g = g + del;      
  }
  return(g);
}
/************************************************************

  get_g(X)

************************************************************/
static double get_g(X)
double X;
{
  static int First_Time = 1;
  static double gamma[NG+1], gprime[NG+1],dN;
  static double a[NG+1], b[NG+1], c[NG+1], d[NG+1];
  int i,j;
  double x, g, di;

  if (First_Time){
    dN = (double)NG;
    for(i=0; i<=NG; i++){
      x = (double)i/dN;
      gamma[i] = g = get_gamma(x);
      gprime[i] = (g < EPS_MIN)?
	3.0:
	(g < (1.0-EPS_MIN))?
	(g*(1-g*g)/(g-2*x+g*x)):
	0.0;
      gprime[i] /= 2*dN;
    }
    for(i=0; i<NG; i++){
      j = i+1;
      a[i] = .25*(gprime[i]+gprime[j])+.25*(gamma[i]-gamma[j]);
      b[i] = .25*(gprime[j]-gprime[i]);
      c[i] = -.25*(gprime[i]+gprime[j])-.75*(gamma[i]-gamma[j]);
      d[i] = .5*(gamma[i]+gamma[j])+.25*(gprime[i]-gprime[j]);
    }
    First_Time = 0;
  }
  i = (int)(dN*X);
  if (i >= NG){
    g = 1.0;
    return(g);
  }
  j = i+1;
  x = (X+X)*dN-(double)(i+j);
  g = d[i]+x*(c[i]+x*(b[i]+x*a[i]));
  return(g);
}

/************************************************************

  get_step_control()

  gets time step control parameters
  mindr_v and mindu_v to use for approximate
  time step control. 

************************************************************/
void get_step_control()
{
unsigned int i, j;
double c, du, vj, maxv_dr, maxv_du, mindr;

/*
  get minimum dr, heavy particle speed, Courant condition on ion
  sound speed
*/
mindr = R;
for(j=1; j<=Nr; j++){
  mindr = min(mindr, (axr->r[j*axr_factor]-axr->r[(j-1)*axr_factor]));
}

maxv_dr = maxv_du = 0.0;
for(i=1; i<=Nu; i++){
  vj = v[j];
  maxv_dr = max(maxv_dr, vj/mindr);
  du = axu->r[i*axu_factor] - axu->r[(i-1)*axu_factor];
  maxv_du = max(maxv_du, vj/du);
}
c = sqrt((2.0*k*Ti)/m_N);
dtkinetic = mindr/c;
dtkinetic = min(dtkinetic, sqrt_3/maxv_dr);
mindu_v = 1.0/maxv_du;
ion_a1 = qi_M/mindr;
ion_a2 = SQR(c/mindr);
return;
}

/************************************************************

  do_wall_absorption()

  With the wall absorption rate determined from Apply_inelastics,
  the pancake functions can be evaluated and the absorption terms
  applied.

  Evaluation of the pancake functions is complicated by the fact
  that logarithmic singularities exist in their formulae. Limit
  approximations are then needed to treat the extreme cases. 

  Result is the wall absorption rate integrated over energy, 
  to give the electron rate of absorption.

************************************************************/
double do_wall_absorption(Xr, Xz, eta, dt)
dblptr Xr, Xz, eta;
double dt;
{
  unsigned int j, pos, posu0, posu1, pos_guard;
  double Xr_, Xz_, eta_, X, Xs, Gr, Gz;
  double g, gr, gz, y, t1, t2, t3, s, cos_t, sin_t;
  double zeta, zGr, zGz, dtrate, eta_incr, result;
  enum g_region {normal, r_axis, z_axis, edge_normal, edge_zaxis, center};
  typedef enum g_region region;
  region current_point;
  bool limit;

  result = 0.0;
  posu0 = Nr*axr_factor;
  posu1 = posu0 + axu_factor;
  for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
    pos_guard = pos + axr_factor;
    Xr_ = Xr[pos]; Xz_ = Xz[pos]; eta_ = eta[pos];
    Xs = Xr_*Xr_+Xz_*Xz_;
    X = sqrt(Xs);
    cos_t = gr = Xr_/X; sin_t = gz = Xz_/X;
    X = (X>1.0)?1.0:X;
    g = get_g(X);
    gr *= g; gz *= g;
    s = sqrt(1.0-gz*gz);
    current_point = normal;
    if (g >= (1.0-EPS_MIN)) {
      if (gr*gr > EPS_ANG){
	current_point = edge_normal;
      } else {
	current_point = edge_zaxis;
      }
    } else if (g <= EPS_MIN) {
      current_point = center;
    } else if (sin_t*sin_t <= EPS_ANG) {
      current_point = r_axis;
    } else if (cos_t*cos_t <= EPS_ANG) {
      current_point = z_axis;
    }
    switch (current_point){
    case normal:
      t1 = log(((g-gr)*(g*s+gr))/((g+gr)*(g*s-gr)));
      zeta = ((1-g*X)/(g*g))*
	((1-s)+
	 .5*(gr/g)*t1
	 );
      zGr = ((1-g*X)/(g*X))*
	(1.5*(1-s)/(g*g)-0.5 +
	 (0.5*gz*gz+1.5*gr*gr/(g*g)-0.5)*t1
	 /(2*g*gr)
	 );
      zGz = ((1-g*X)/(g*X))*
	((g*g-1.5*gr*gr)/(g*g*(1+s)) - 0.5 + 
	 .25*(gr/g)*(3.0/(g*g)-1)*t1
	 );
      break;
    case edge_normal:
      y = fabs(gr/g);
      zeta = ((1-g*X)*(1-s+y*(t1=log((s+y)/((1+y)*(1+g))))) + y*g)/(g*g);
      t2 = (gz*gz+3*y*y-1)/(g*g*y);
      zGr = ((1-g*X)*(3*(1-s)/(g*g)-1+t2*t1))/(2*g*X)+t2/(2*X);
      t3 = y*(3/(g*g)-1);
      zGz = ((1-g*X)*((2-3*y*y)/(1+s)-1+t3*t1)/(2*g*X)) + t3/(2*X);
      break;
    case edge_zaxis:
      zeta = (1-s)*(1-g*X)/(g*g);
      zGr = .5*((1-g*X)/(g*X))*((1-s)*(3-s)/(g*g)-1.0);
      zGz = .5*((1-g*X)/(g*X))*((1-s)/(1+s));
      break;
    case center:
      zeta = 0.5;
      zGr = 0.75;
      zGz = 0.375;
      break;
    case r_axis:
      zeta = (log(1-g*g)/(log(1-g)-log(1+g)))/g;
      zGr = -.5*((1-g*X)/(g*X))*(g*g + log(1-g*g))/(g*g);
      zGz = -.5*((1-g*X)/(g*X))*(1.5+.5*(3/(g*g)-1.0)*log(1-g*g));
      break;
    case z_axis:
      zeta = (1-s)*(1-g*X)/(g*g);
      zGr = .5*((1-g*X)/(g*X))*((1-s)*(3*s+(g*g-1))/(g*g*s)-1.0);
      zGz = .5*((1-g*X)/(g*X))*((1-s)/(1+s));
      break;
    }
    dtrate = dt*wall_absorption_rate[j];
    eta_incr = 1.0+dtrate*zeta;
    eta[pos_guard] = SBC[j]*(eta[pos] = eta_/eta_incr);
    result -= eta[pos]*wall_absorption_rate[j]*zeta*axu->Volume[pos];
    Xr[pos_guard] = (Xr[pos] = Xr_*(eta_incr/(1.0+dtrate*zGr)));
    Xz[pos_guard] = (Xz[pos] = Xz_*(eta_incr/(1.0+dtrate*zGz)));
  }
  result *= EOM;
  return(result);
}
/************************************************************

  get_elastic()

  fills arrays elastic and del_elastic (corrected for heavy temp)
  and similar arrays.

************************************************************/
void get_elastic()
{
unsigned int j;
double u, fac1;

elastic = get_memory(elastic, 3*(Nu+2), double);
del_elastic = elastic + (Nu+2);
u_diff = del_elastic + (Nu+2);

 for(j=0; j<=(Nu+1); j++){
   u = j?axu->r[j*axu_factor]:0.0;
   elastic[j] = q_mom(u)*v[j];
   del_elastic[j] = delel*elastic[j];
 }
 for(j=0; j<=Nu; j++){
   /*   u = j?axu->r[j*axu_factor]:0.0;*/
   u = axu->r[j*axu_factor];
   u = .5*(u + axu->r[(j+1)*axu_factor]);
   u = j?u:0.0;
   u_diff[j] = TWO_THIRDS*delel*q_mom(u)*STEOM*u*sqrt(u);
 }
 u_diff[Nu+1] = u_diff[Nu];
 
del_elastic[0] = 0.0;
return;
}
/************************************************************

  get_inelastics()

  gets arrays inelastic, ionization, and de_excitation_array
  which describe inelastic processes and re-distribution of
  de-excited electrons.

  ionization energy loss is shared between the two
  resultant electrons.

  de_excitation_dat contains pointers to pointers to arrays
  of collision values and origins for the inelastic terms.
  de_excitation_dat[i].val[j] is the value at energy i due to
  de_excitation from energy at de_excitation_dat[i].orig[j]. 
  de_excitation_dat[i].num contains the number of such levels
  (values of j) which contribute to level i.  

  Wall recombination rate is established as v/(2*dr) where
  dr is the distance between the last cell and the guard cell. 

************************************************************/
void get_inelastics(set)
MD_FCT_species_set_ptr set;
{
unsigned int i, j;
double fac1;
static dblptr de_excitation_array = NULL, *de_excitation_ptr = NULL;
MD_FCT_species_ptr sp;
static string gstring = NULL;
/*
  get data space 
*/
gstring = get_memory(gstring, 50, char);
inelastic = get_memory(inelastic, (Nr+2)*(Nu+2), double);
wall_absorption_rate = get_memory(wall_absorption_rate, (Nu+2), double);
de_excitation_array = get_memory(de_excitation_array, ((Nu+2)*(Nu+2)), double);
de_excitation_ptr = get_memory(de_excitation_ptr, (Nu+2), dblptr);

excitation_events = ionization_events = NULL;
excitation_events = get_memory(excitation_events, Nex, inelastic_event);
ionization_events = get_memory(ionization_events, Nion, inelastic_event);
/*
  allocate space for inelastic events
*/
 for(i=0; i<Nex; i++){
   excitation_events[i].multiplicity = 1.0;
   excitation_events[i].frequency = NULL;
   excitation_events[i].jmin = NULL;
   sp = NULL;
   if (Ex_targets[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ex_targets[i]);
     sp = MD_get_spptr(gstring, set);
   }
   excitation_events[i].target = sp;
   sp = NULL;
   if (Ex_products[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ex_products[i]);
     sp = MD_get_spptr(gstring, set);
   }
   excitation_events[i].product = sp;
   excitation_events[i].frequency = get_memory(excitation_events[i].frequency, 3*(Nu+2), double);
   excitation_events[i].lambda_min = excitation_events[i].frequency + (Nu+2);
   excitation_events[i].lambda_max = excitation_events[i].lambda_min + (Nu+2);
   excitation_events[i].jmin = get_memory(excitation_events[i].jmin, 2*(Nu+2), unsigned int);
   excitation_events[i].jmax = excitation_events[i].jmin + (Nu+2);
   excitation_events[i].loss_rate = Ex_loss_rates[i];
   for(j=0; j<=Nu+1; j++){
     excitation_events[i].lambda_min[j] = excitation_events[i].lambda_max[j] = 0.0;
     excitation_events[i].jmin[j] = excitation_events[i].jmax[j] = 0;
   }
 }
 for(i=0; i<Nion; i++){
   sp = NULL;
   if (Ion_targets[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ion_targets[i]);
     sp =  MD_get_spptr(gstring, set);
   }
   ionization_events[i].target =sp;
   sp = NULL;
   if (Ion_products[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ion_products[i]);
     sp = MD_get_spptr(gstring, set);
   }
   ionization_events[i].product = sp;
   ionization_events[i].multiplicity = 2.0;
   ionization_events[i].frequency = NULL;
   ionization_events[i].jmin = NULL;
   ionization_events[i].frequency = get_memory(ionization_events[i].frequency, 3*(Nu+2), double);
   ionization_events[i].lambda_min = ionization_events[i].frequency + (Nu+2);
   ionization_events[i].lambda_max = ionization_events[i].lambda_min + (Nu+2);
   ionization_events[i].jmin = get_memory(ionization_events[i].jmin, 2*(Nu+2), unsigned int);
   ionization_events[i].jmax = ionization_events[i].jmin + (Nu+2);
   ionization_events[i].loss_rate = 0.0;
   for(j=0; j<=Nu+1; j++){
     ionization_events[i].lambda_min[j] = ionization_events[i].lambda_max[j] = 0.0;
     ionization_events[i].jmin[j] = ionization_events[i].jmax[j] = 0;
   }

 }
 /* pack the frequency arrays */
for(i=0; i<=(Nu+1); i++){
  fac1 = v[i];
  for(j=0; j<Nion; j++){
    ionization_events[j].frequency[i] = fac1*Q_ion[j](u[i]);
    de_excitation(*E_ion[j], axu, i, ionization_events[j]);
  }
  for(j=0; j<Nex; j++){
    excitation_events[j].frequency[i] = fac1*Q_ex[j](u[i]);
    de_excitation(*E_ex[j], axu, i, excitation_events[j]);
  }
}
/*
  get space for the wall collision parameters...E3
*/
 SBC = get_memory(SBC, (Nu+2), double);
return;
}
/************************************************************

  de_excitation()

  interpolates an energy axis to find where to put
  de-excitation rate at energy upp. Interpolates
  between nearest destinations.

  Linear interpolates:

     ax_ptr->r[pp] ...  upp  ...  ax_ptr->r[pp+1]
         ^                            ^
         |                            |
      (1 - lambda)                 lambda

  where lambda = (upp - r[pp])/(r[pp+1] - r[pp]).

  If not in range, makes sure lambda = 0.0 and pp = 0
  and returns false.

  Then it distributes the de-excited electrons in
  energy cells according to the relative sizes
  of the original cell and destination cells.

  del_E is the energy loss involved
  ax_ptr is axis pointer for energy axis
  nu is de-excitation rate 
  orig is the index of the originating energy cell
  event is the inelastic event
  (energy is shared equally among multiple electrons).
  
************************************************************/
static bool de_excitation(del_E, ax_ptr, orig, event)
     double del_E;
     axis_ptr ax_ptr;
     unsigned int orig;
     inelastic_event event;
{
  unsigned int pp, i, N, j, pos, pos1, ppos, ppos1, origpos, jpos, jj;
  double lambda, VL, VH, Vi, Vj, Vpp, Vppp, v, upp, x, u;
  double multiplicity;
  dblptr r;
  bool found, not_done, offscale;
  
  multiplicity = event.multiplicity;
  origpos = orig*axu_factor;
  r = ax_ptr->r;
  u = r[origpos];
  upp = (u - del_E)/multiplicity;
  N = ax_ptr->num_pts + 1;
  
  found = false;
  for(i=0; (!found)&&(i<N); i++){
    pos = i*axu_factor;
    pos1 = pos + axu_factor;
    found = ((r[pos] <= upp) && (upp < r[pos1]));
    pp = i;
  }
  ppos = pp*axu_factor;
  ppos1 = ppos+axu_factor;
  offscale = false;
  if (found){
    lambda = (upp - r[ppos])/(r[ppos1] - r[ppos]);
  } else {
    lambda = 0.0;
    pp = 0;
    ppos = 0;
    ppos1 = 0;
  }
  if (upp > ax_ptr->r[N*axu_factor]){
    VL = (upp - (ax_ptr->r[N] + .5*ax_ptr->Volume[N]));
    if (VL > 0.0){
      pp = N;
      ppos = N*axu_factor;
      ppos1 = N*axu_factor;
      found = true;
      offscale = true;
    }
  }
  if (found) {
    Vi = ax_ptr->Volume[origpos]/multiplicity;
    Vpp = ax_ptr->Volume[pp?(pp*axu_factor):axu_factor];
    Vppp = ax_ptr->Volume[(pp+1)*axu_factor];
    if (lambda <= .5){
      VL = min(Vi,(.5*Vi + (.5-lambda)*Vpp));
      VH = (Vi - VL);
    } else {
      VH = min(Vi,(.5*Vi - (.5-lambda)*Vppp));
      VL = Vi-VH;
    }
    if (offscale){
      VH = 0.0;
      VL = upp - (ax_ptr->r[N] + .5*ax_ptr->Volume[N]);
    }
    if (pp == 0) {VL = 0.0;}
    if (pp == 1) {VL = ax_ptr->Volume[axu_factor];}
    v = VL; j = pp?pp:1;
    do {
      jpos = j*axu_factor;
      Vj = ax_ptr->Volume[jpos];
      not_done = ((v > Vj)&&(j != 1));
      if (not_done){
	v -= Vj;
      } else {
	x = (v/Vj);
	jj = j;
      }
      j--;
    } while(not_done);
    event.jmin[orig] = pp?jj:0;
    event.lambda_min[orig] = x;
    if (!offscale){
      v = VH; j = (pp < N)?pp+1:N;
      do {
	jpos = j*axu_factor;
	Vj = ax_ptr->Volume[jpos];
	not_done = ((v > Vj)&&(j < N));
	if (not_done){
	  v -= Vj;
	} else {
	  x = (v/Vj);
	  jj = j;
	}
	j++;
      } while(not_done);
      event.jmax[orig] = jj;
      event.lambda_max[orig] = x;
    }
  }
  return(found);
}
/************************************************************

  Apply_inelastics()

  Applies each inelastic event to all electrons, and
  updates targets and products.
  Also obtains one_elastic.
  
  target_first: first fractional step for targets, take
  rho_n -> rho_np1; use rho_n for target.
  otherwise, take rho_np1->rho_np1;

  product_first: first fractional step, take rho_n -> rho_np1.
  otherwise take rho_np1 -> rho_np1.

  Also updates inelastics, as these are functions of all
  phase space.

************************************************************/
void Apply_inelastics(rho, dt, target_first, product_first, Xterm, nN_n, uN_n)
dblptr rho, Xterm, nN_n, uN_n;
double dt;
bool  target_first, product_first;
{
static dblptr temp = NULL, incr = NULL;
double nu, etabc_r_0, rho_target, increment, d, vj, psi, E3; 
double destruction, nup, f, Vol, nueta, nuM, nN_, uN;
unsigned int i, j, p,l, pos, posr, posu, posu0, posu1, start, stop;
static bool First_Time = true;
inelastic_event_ptr event;
MD_FCT_species_ptr sp;
static unsigned int r_R, rusize, u_infty;

if (First_Time){
  r_R = (Nr+1)*axr_factor;
  rusize = (Nr+2)*(Nu+2);
  u_infty = (Nu+1)*axu_factor;

  temp = get_memory(temp, (Nu+2), double);
  incr = get_memory(incr, (Nu+2), double);
  First_Time = false;
}
/*
  As rates at a given r are collected, apply energy loss/gain term
  to rho. Then, apply excitation/ionization depletion for all
  inelastic events.

  Finally, go through and increment all products; decrement all targets;
  according to calculated rates.

*/
 d = R - axr->r[axr_factor*Nr];
 for(i = 1; i<=Nr; i++){
   posu0 = i*axr_factor;
   posu1 = posu0+axu_factor;
   nN_ = nN_n[i]; uN = ONE_THIRD*uN_n[i];
   for(l=0; l<Nex; l++){
     event = (excitation_events+l);
     event->incr = event->decr = 0.0;
     if (target_first){
       event->target->rho_np1[i] = event->target->rho_n[i];
     }
     if ((product_first)&&(event->product != NULL)) {
       event->product->rho_np1[i] = event->product->rho_n[i];
     }
   }
   for(l=0; l<Nion; l++){
     event = (ionization_events+l);
     event->incr = event->decr = 0.0;
     if ((event->product != NULL) && (product_first)){
       event->product->rho_np1[i] = event->product->rho_n[i];
     }
   }
   for(j=0; j<=(Nu+1); j++){
     incr[j] = 0.0;
   }
   for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
     inelastic[pos] = 0.0;
     Vol = axu->Volume[pos];
     for(l=0; l<Nex; l++){
       event = (excitation_events+l);
       rho_target = event->target->rho_np1[i];
       f = event->frequency[j];
       nu = f*rho_target;
       nup = f*rho[pos];
       nueta = nup*rho_target;
       event->incr += nueta*Vol;
       event->decr += nup*Vol;
       start = event->jmin[j]; 
       stop = event->jmax[j];
       if (stop != 0){
	 for(p=start+1; p<stop; p++){
	   incr[p] += nueta;
	 }
	 incr[start] += nueta*(event->lambda_min[j]);
	 incr[stop] += nueta*(event->lambda_max[j]);
       }
       inelastic[pos] += nu;
     }
     for(l=0; l<Nion; l++){
       event = (ionization_events+l);
       rho_target = event->target->rho_np1[i];
       f = event->frequency[j];
       nu = f*rho_target;
       nup = f*rho[pos];
       nueta = nup*rho_target;
       event->incr += nueta*Vol;
       event->decr += nup*Vol;
       nueta = nueta*(event->multiplicity)*(event->multiplicity);
       start = event->jmin[j]; 
       stop = event->jmax[j]; 
       if (stop != 0){
	 for(p=start+1; p<stop; p++){
	   incr[p] += nueta;
	 }
	 incr[start] += nueta*(event->lambda_min[j]);
	 incr[stop] += nueta*(event->lambda_max[j]);
       }
       inelastic[pos] += nu;
     }
     nuM = nN_*elastic[j]+inelastic[pos];
     if (i == Nr){
       vj = v[j];
       psi = d*nuM/vj;
       E3 = En(psi, 3);
       wall_absorption_rate[j] = E3*vj/d;
       SBC[j] = 4.0*E3-1.0;
     }
   }
   /* 
      update all the participants
   */
   for(l=0; l<Nex; l++){
     event = (excitation_events + l);
     increment = dt*event->incr*EOM;
     destruction = 1.0/(1.0 + dt*event->decr*EOM);
     event->target->rho_np1[i] *=  destruction;
     if (event->product != NULL){
       destruction = 1.0/(1.0+dt*event->loss_rate);
       event->product->rho_np1[i] += increment;
       event->product->rho_np1[i] *= destruction;
     }
   }
   for(l=0; l<Nion; l++){
     event = (ionization_events + l);
     increment = dt*event->incr*EOM;
     destruction = 1.0/(1.0 + dt*event->decr*EOM);
     event->target->rho_np1[i] *=  destruction;
     if (event->product != NULL){
       event->product->rho_np1[i] += increment;
     }
   }
   /*nuM = nN_*elastic[j]+inelastic[pos];*/
   for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
     Xterm[pos] -= (incr[j]/rho[pos]) + 
       (
	   (
	    (nN_*u[j+1]-uN)*del_elastic[j+1]-(nN_*u[j-1]-uN)*del_elastic[j-1]
	    )*one_del_2u[j] + 
	   elastic[j]*nN_
	   );
     rho[pos] = (rho[pos] + dt*incr[j])/(1.0+dt*inelastic[pos]);
   }
 }
 /*
   Do boundary conditions along u=0, u=infty
 */
 pos = 0;
 for(i=1; i<=Nr; i++){
   pos += axr_factor;
   posu = pos + u_infty;
   rho[pos] = (etasp->bc[axu->sequence].Sl)*rho[pos+axu_factor];
   rho[posu] = eta_TBC[i]*rho[posu-axu_factor];
 }
 /*
   Then along r=0, r=R and fix up corners
 */
 pos = 0;
 etabc_r_0 = (etasp->bc[axr->sequence].Sl);
 SBC[Nu+1] = SBC[Nu]; SBC[0] = 0.0;
 for(j=0; j<=(Nu+1); j++){
   posr = pos+r_R - axr_factor;
   rho[pos] = etabc_r_0*rho[pos+axr_factor];
   rho[posr+axr_factor] = SBC[j]*rho[posr];
   pos += axu_factor;
 }
 for(l=0; l<Nex; l++){
   event = (excitation_events + l);
   sp = event->target; 
   sp->rho_np1[0] = sp->bc[axr_config->sequence].Sl*sp->rho_np1[1];
   sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
   if (event->product != NULL){
     sp = event->product; 
     sp->rho_np1[0] = sp->bc[axr_config->sequence].Sl*sp->rho_np1[1];
     sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
   }
 }
 for(l=0; l<Nion; l++){
   event = (ionization_events + l);
   sp = event->target; 
   sp->rho_np1[0] = sp->bc[axr_config->sequence].Sl*sp->rho_np1[1];
   sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
   if (event->product != NULL){
     sp = event->product; 
     sp->rho_np1[0] = sp->bc[axr_config->sequence].Sl*sp->rho_np1[1];
     sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
   }
 }  
 return;
}
