/************************************************************

  collisions.c

  Various routines for acquiring collision data and 
  arranging it into various arrays. 
  Also has routine for applying inelastic collision term.
  Part of "poscol" simulation.

************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "basics.h"
#undef pi
#include "MD_FCT_defs.h"
#include "MD_FCT.h"
#include "MD_FCT_utils.h"
#include "poscol.h"
#include "sources.h"
#include "special.h"

typedef struct inelastic_event{
  MD_FCT_species_ptr target;
  MD_FCT_species_ptr product;
  double multiplicity;
  dblptr frequency;
  unsigned_int_ptr jmin;
  dblptr lambda_min;
  unsigned_int_ptr jmax;
  dblptr lambda_max;
  double loss_rate;
  double incr;
  double decr;
} inelastic_event;

typedef inelastic_event *inelastic_event_ptr;

static bool de_excitation(double, axis_ptr, unsigned int, inelastic_event);
static dblptr inelastic = NULL, ionization = NULL;
dblptr SBC = NULL, alpha = NULL, beta = NULL;
static inelastic_event_ptr excitation_events = NULL, ionization_events = NULL;

/************************************************************

  get_step_control()

  gets time step control parameters
  mindr_v and mindu_v to use for approximate
  time step control. 

************************************************************/
void get_step_control()
{
unsigned int i, j;
double b, c, du, v, maxv_dr, maxv_du, mindr, tmpdt;

/*
  get minimum dr, heavy particle speed, Courant condition on ion
  sound speed
*/
alpha = get_memory(alpha, (Nu+2), double);
beta = get_memory(beta, (Nu+2), double);
mindr = R;
for(j=1; j<=Nr; j++){
  mindr = min(mindr, (axr->r[j*axr_factor]-axr->r[(j-1)*axr_factor]));
}
c = sqrt((2.0*k*Ti)/m_N);
dtkinetic = mindr/c;
maxv_dr = maxv_du = 0.0;
for(i=1; i<=Nu; i++){
  v = ve[i];
  c = mindr*mindr*3.0/(v*v);
  b = beta[i] = (elastic[i]+inelastic[i])*c;
  tmpdt = .5*(b+sqrt(b*b+4.0*c));
  dtkinetic = min(dtkinetic, tmpdt);

  du = axu->r[i*axu_factor] - axu->r[(i-1)*axu_factor];
  alpha[i] = (v*v)*(2.0/3.0)/(du*du);
}
ion_a1 = qi_M/mindr;
ion_a2 = SQR(c/mindr);
return;
}
/************************************************************

  get_elastic()

  fills arrays elastic and del_elastic

  also gets u_h (= sqrt(u)) 

************************************************************/
void get_elastic()
{
unsigned int j;
double u, fac1;

elastic = get_memory(elastic, 6*(Nu+2), double);
del_elastic = elastic + (Nu+2);
u_h =  del_elastic + (Nu+2);
one_u_h =  u_h + (Nu+2);
ve = one_u_h + (Nu+2);
vs = ve + (Nu+2);

for(j=0; j<=(Nu+1); j++){
  u = j?axu->r[j*axu_factor]:0.0;
  vs[j] = teom*u;
  u_h[j] = sqrt(u);
  ve[j] = steom*u_h[j];
  one_u_h[j] = 1.0/(j?u_h[j]:1.0);
  fac1 = ve[j]*nN;
  elastic[j] = q_mom(u)*fac1;
  del_elastic[j] = delel*elastic[j]*u;
}
one_u_h[0] = 1.0/u_h[1];
del_elastic[0] = 0.0;
vs[0] = vs[1];
ve[0] = vs[1];
return;
}
/************************************************************
  
  Du()

  gets diffusion coefficients for u-diffusion term

************************************************************/
MD_diffusivity Du(spe, pos, u)
MD_FCT_species_ptr spe;
unsigned int pos;
double u;
{
MD_diffusivity result;

if (u>0.0){
  result.D = delel*steom*nN*kTi_e*u*u*q_mom(u);
  result.kappa = 1.0/sqrt(u);
} else {
  result.D = 0.0; result.kappa = 1.0;
}
return(result);
}
/************************************************************

  get_inelastics()

  gets arrays inelastic, ionization, and de_excitation_array
  which describe inelastic processes and re-distribution of
  de-excited electrons.

  ionization energy loss is shared between the two
  resultant electrons.

  de_excitation_dat contains pointers to pointers to arrays
  of collision values and origins for the inelastic terms.
  de_excitation_dat[i].val[j] is the value at energy i due to
  de_excitation from energy at de_excitation_dat[i].orig[j]. 
  de_excitation_dat[i].num contains the number of such levels
  (values of j) which contribute to level i.  

  Wall recombination rate is established as v/(2*dr) where
  dr is the distance between the last cell and the guard cell. 

************************************************************/
void get_inelastics(set)
MD_FCT_species_set_ptr set;
{
unsigned int i, j;
double u, v, fac1, rho_target, nu;
static dblptr de_excitation_array = NULL, *de_excitation_ptr = NULL;
MD_FCT_species_ptr sp;
static string gstring = NULL;
/*
  get data space 
*/
gstring = get_memory(gstring, 50, char);
inelastic = get_memory(inelastic, (Nu+2), double);
ionization = get_memory(ionization, (Nu+2), double);
nuM = get_memory(nuM, (Nr+2)*(Nu+2), double);
de_excitation_array = get_memory(de_excitation_array, ((Nu+2)*(Nu+2)), double);
de_excitation_ptr = get_memory(de_excitation_ptr, (Nu+2), dblptr);

excitation_events = ionization_events = NULL;
excitation_events = get_memory(excitation_events, Nex, inelastic_event);
ionization_events = get_memory(ionization_events, Nion, inelastic_event);
/*
  allocate space for inelastic events
*/
 for(i=0; i<Nex; i++){
   excitation_events[i].multiplicity = 1.0;
   excitation_events[i].frequency = NULL;
   excitation_events[i].jmin = NULL;
   sp = NULL;
   if (Ex_targets[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ex_targets[i]);
     sp = MD_get_spptr(gstring, set);
   }
   excitation_events[i].target = sp;
   sp = NULL;
   if (Ex_products[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ex_products[i]);
     sp = MD_get_spptr(gstring, set);
   }
   excitation_events[i].product = sp;
   excitation_events[i].frequency = get_memory(excitation_events[i].frequency, 3*(Nu+2), double);
   excitation_events[i].lambda_min = excitation_events[i].frequency + (Nu+2);
   excitation_events[i].lambda_max = excitation_events[i].lambda_min + (Nu+2);
   excitation_events[i].jmin = get_memory(excitation_events[i].jmin, 2*(Nu+2), unsigned int);
   excitation_events[i].jmax = excitation_events[i].jmin + (Nu+2);
   excitation_events[i].loss_rate = Ex_loss_rates[i];
   for(j=0; j<=Nu+1; j++){
     excitation_events[i].lambda_min[j] = excitation_events[i].lambda_max[j] = 0.0;
     excitation_events[i].jmin[j] = excitation_events[i].jmax[j] = 0;
   }
 }
 for(i=0; i<Nion; i++){
   sp = NULL;
   if (Ion_targets[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ion_targets[i]);
     sp =  MD_get_spptr(gstring, set);
   }
   ionization_events[i].target =sp;
   sp = NULL;
   if (Ion_products[i] != NULL){
     gstring = strcpy(gstring, "n");
     gstring = strcat(gstring, Ion_products[i]);
     sp = MD_get_spptr(gstring, set);
   }
   ionization_events[i].product = sp;
   ionization_events[i].multiplicity = 2.0;
   ionization_events[i].frequency = NULL;
   ionization_events[i].jmin = NULL;
   ionization_events[i].frequency = get_memory(ionization_events[i].frequency, 3*(Nu+2), double);
   ionization_events[i].lambda_min = ionization_events[i].frequency + (Nu+2);
   ionization_events[i].lambda_max = ionization_events[i].lambda_min + (Nu+2);
   ionization_events[i].jmin = get_memory(ionization_events[i].jmin, 2*(Nu+2), unsigned int);
   ionization_events[i].jmax = ionization_events[i].jmin + (Nu+2);
   ionization_events[i].loss_rate = 0.0;
   for(j=0; j<=Nu+1; j++){
     ionization_events[i].lambda_min[j] = ionization_events[i].lambda_max[j] = 0.0;
     ionization_events[i].jmin[j] = ionization_events[i].jmax[j] = 0;
   }

 }
 /* pack the frequency arrays */
for(i=0; i<=(Nu+1); i++){
  v = ve[i];
  fac1 = v;
  u = axu->r[i*axu_factor];
  ionization[i] = inelastic[i] = 0.0;
  for(j=0; j<Nion; j++){
    rho_target = ionization_events[j].target->rho0[0];
    nu = ionization_events[j].frequency[i] = fac1*Q_ion[j](u)*rho_target;
    de_excitation(*E_ion[j], axu, i, ionization_events[j]);
    inelastic[i] += nu;
    ionization[i] += nu;
  }
  for(j=0; j<Nex; j++){
    rho_target = excitation_events[j].target->rho0[0];
    nu = excitation_events[j].frequency[i] = fac1*Q_ex[j](u)*rho_target;
    de_excitation(*E_ex[j], axu, i, excitation_events[j]);
    inelastic[i] += nu;
  }
}
/*
  get space for the wall collision parameters...E3
*/
SBC = get_memory(SBC, (Nu+2), double);

return;
}
/************************************************************

  de_excitation()

  interpolates an energy axis to find where to put
  de-excitation rate at energy upp. Interpolates
  between nearest destinations.

  Linear interpolates:

     ax_ptr->r[pp] ...  upp  ...  ax_ptr->r[pp+1]
         ^                            ^
         |                            |
      (1 - lambda)                 lambda

  where lambda = (upp - r[pp])/(r[pp+1] - r[pp]).

  If not in range, makes sure lambda = 0.0 and pp = 0
  and returns false.

  Then it distributes the de-excited electrons in
  energy cells according to the relative sizes
  of the original cell and destination cells.

  del_E is the energy loss involved
  ax_ptr is axis pointer for energy axis
  nu is de-excitation rate 
  orig is the index of the originating energy cell
  event is the inelastic event
  (energy is shared equally among multiple electrons).
  
************************************************************/
static bool de_excitation(del_E, ax_ptr, orig, event)
     double del_E;
     axis_ptr ax_ptr;
     unsigned int orig;
     inelastic_event event;
{
  unsigned int pp, i, N, j, pos, pos1, ppos, ppos1, origpos, jpos, jj;
  double lambda, VL, VH, Vi, Vj, Vpp, Vppp, v, upp, x, u;
  double multiplicity;
  dblptr r;
  bool found, not_done, offscale;
  
  multiplicity = event.multiplicity;
  origpos = orig*axu_factor;
  r = ax_ptr->r;
  u = r[origpos];
  upp = (u - del_E)/multiplicity;
  N = ax_ptr->num_pts + 1;
  
  found = false;
  for(i=0; (!found)&&(i<N); i++){
    pos = i*axu_factor;
    pos1 = pos + axu_factor;
    found = ((r[pos] <= upp) && (upp < r[pos1]));
    pp = i;
  }
  ppos = pp*axu_factor;
  ppos1 = ppos+axu_factor;
  offscale = false;
  if (found){
    lambda = (upp - r[ppos])/(r[ppos1] - r[ppos]);
  } else {
    lambda = 0.0;
    pp = 0;
    ppos = 0;
    ppos1 = 0;
  }
  if (upp > ax_ptr->r[N*axu_factor]){
    VL = (upp - (ax_ptr->r[N] + .5*ax_ptr->Volume[N]));
    if (VL > 0.0){
      pp = N;
      ppos = N*axu_factor;
      ppos1 = N*axu_factor;
      found = true;
      offscale = true;
    }
  }
  if (found) {
    Vi = ax_ptr->Volume[origpos]/multiplicity;
    Vpp = ax_ptr->Volume[pp?(pp*axu_factor):axu_factor];
    Vppp = ax_ptr->Volume[(pp+1)*axu_factor];
    if (lambda <= .5){
      VL = min(Vi,(.5*Vi + (.5-lambda)*Vpp));
      VH = (Vi - VL);
    } else {
      VH = min(Vi,(.5*Vi - (.5-lambda)*Vppp));
      VL = Vi-VH;
    }
    if (offscale){
      VH = 0.0;
      VL = upp - (ax_ptr->r[N] + .5*ax_ptr->Volume[N]);
    }
    if (pp == 0) {VL = 0.0;}
    if (pp == 1) {VL = ax_ptr->Volume[axu_factor];}
    v = VL; j = pp?pp:1;
    do {
      jpos = j*axu_factor;
      Vj = ax_ptr->Volume[jpos];
      not_done = ((v > Vj)&&(j != 1));
      if (not_done){
	v -= Vj;
      } else {
	x = (v/Vj);
	jj = j;
      }
      j--;
    } while(not_done);
    event.jmin[orig] = pp?jj:0;
    event.lambda_min[orig] = x;
    if (!offscale){
      v = VH; j = (pp < N)?pp+1:N;
      do {
	jpos = j*axu_factor;
	Vj = ax_ptr->Volume[jpos];
	not_done = ((v > Vj)&&(j < N));
	if (not_done){
	  v -= Vj;
	} else {
	  x = (v/Vj);
	  jj = j;
	}
	j++;
      } while(not_done);
      event.jmax[orig] = jj;
      event.lambda_max[orig] = x;
    }
  }
  return(found);
}
/************************************************************

  Apply_inelastics()

  Applies each inelastic event to all electrons.
  
  target_first: first fractional step for targets, take
  rho_n -> rho_np1; use rho_n for target.
  otherwise, take rho_np1->rho_np1;

  product_first: first fractional step, take rho_n -> rho_np1.
  otherwise take rho_np1 -> rho_np1.

  Also updates nuM and inelastics, as these are functions of all
  phase space.

************************************************************/
void Apply_inelastics(rho, dt, target_first, product_first)
dblptr rho;
double dt;
bool  target_first, product_first;
{
  static dblptr loss_rate = NULL, incr = NULL;
  double nu, etabc_r_0, rho_target, increment, d, v, psi, E3, rhop; 
  double destruction, nup, f, Vol, nueta, wall_rate, dtFOURPIEOM;
  unsigned int i, j, p,l, pos, posr, posu, posu0, posu1, start, stop;
  static bool First_Time = true;
  inelastic_event_ptr event;
  MD_FCT_species_ptr sp;

  if (First_Time){
    loss_rate = get_memory(loss_rate, (Nu+2), double);
    incr = get_memory(incr, (Nu+2), double);
    First_Time = false;
  }
/*
  As rates at a given r are collected, apply energy loss/gain term
  to rho. Then, apply excitation/ionization depletion for all
  inelastic events.

  Finally, Go through and increment all products.

*/
  dtFOURPIEOM = dt*FOURPIEOM;
 d = R - axr->r[axr_factor*Nr];
 for(i = 1; i<=Nr; i++){
   posu0 = i*axr_factor;
   posu1 = posu0+axu_factor;
   for(l=0; l<Nex; l++){
     event = excitation_events+l;
     event->incr = event->decr = 0.0;
     if (target_first){
       event->target->rho_np1[i] = event->target->rho_n[i];
     }
     if ((product_first)&&(event->product != NULL)) {
       event->product->rho_np1[i] = event->product->rho_n[i];
     }
   }
   for(l=0; l<Nion; l++){
     event = ionization_events+l;
     event->incr = event->decr = 0.0;
     if ((event->product != NULL) && (product_first)){
       event->product->rho_np1[i] = event->product->rho_n[i];
     }
   }
   for(j=0; j<=(Nu+1); j++){
     incr[j] = 0.0;
   }
   for(j=1, pos = posu0; j<=Nu; j++, pos += axu_factor){
     Vol = axu->Volume[pos];
     rhop = rho[pos];
     for(l=0; l<Nex; l++){
       event = excitation_events+l;
       nu = f = event->frequency[j];
       nup = f*rhop;
       nueta = nup;
       event->incr += nup*Vol;
       event->decr += nup*Vol;
       start = event->jmin[j]; 
       stop = event->jmax[j];
       if (stop != 0){
	 for(p=start+1; p<stop; p++){
	   incr[p] += nueta;
	 }
	 incr[start] += nueta*(event->lambda_min[j]);
	 incr[stop] += nueta*(event->lambda_max[j]);
       }
     }
     for(l=0; l<Nion; l++){
       event = ionization_events+l;
       nu = f = event->frequency[j];
       nup = f*rhop;
       nueta = nup;
       event->incr += nueta*Vol;
       event->decr += nup*Vol;
       nueta = nueta*(event->multiplicity)*(event->multiplicity);
       start = event->jmin[j]; 
       stop = event->jmax[j]; 
       if (stop != 0){
	 for(p=start+1; p<stop; p++){
	   incr[p] += nueta;
	 }
	 incr[start] += nueta*(event->lambda_min[j]);
	 incr[stop] += nueta*(event->lambda_max[j]);
       }
     }
   /*
     collect the boundary conditions 
   */
     nuM[pos] = elastic[j]+inelastic[j];
     loss_rate[j] = inelastic[j];
     if (i == Nr){
       v = ve[j];
       psi = d*nuM[pos]/v;
       E3 = En(psi, 3);
       wall_rate = E3*v/d;
       loss_rate[j] = inelastic[j] + wall_rate;
       nuM[pos] += wall_rate;
       SBC[j] = 4.0*E3-1.0;
     }
   }
   /* 
      update all the participants
   */
   for(l=0; l<Nex; l++){
     event = excitation_events + l;
     increment = event->incr*dtFOURPIEOM;
     if (event->product != NULL){
       event->product->rho_np1[i] += increment;
     }
   }
   for(l=0; l<Nion; l++){
     event = ionization_events + l;
     increment = event->incr*dtFOURPIEOM;
     if (event->product != NULL){
       event->product->rho_np1[i] += increment;
     }
   }
   for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
     rho[pos] = (rho[pos] + dt*incr[j])/(1.0+dt*loss_rate[j]);
   }
 }
 /*
   Do boundary conditions along u=0, u=infty
 */
 pos = 0;
 for(i=1; i<=Nr; i++){
   pos += axr_factor;
   posu = pos + u_infty;
   rho[pos] = (etasp->bc[axu->sequence].Sl)*rho[pos+axu_factor];
   rho[posu] = eta_TBC[i]*rho[posu-axu_factor];
 }
 /*
   Then along r=0, r=R and fix up corners
 */
 pos = 0;
 etabc_r_0 = (etasp->bc[axr->sequence].Sl);
 SBC[Nu+1] = SBC[Nu]; SBC[0] = 0.0;
 for(j=0; j<=(Nu+1); j++){
   posr = pos+r_R - axr_factor;
   rho[pos] = etabc_r_0*rho[pos+axr_factor];
   rho[posr+axr_factor] = SBC[j]*rho[posr];
   pos += axu_factor;
 }
 for(l=0; l<Nex; l++){
   event = excitation_events + l;
   if (event->product != NULL){
     sp = event->product;
     sp->rho_np1[0] = sp->bc[axr_config->sequence].Sl*sp->rho_np1[1];
     sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
   }
 }
 for(l=0; l<Nion; l++){
   event = ionization_events + l;
   if (event->product != NULL){
     sp = event->product;
     sp->rho_np1[0] = sp->bc[axr_config->sequence].Sl*sp->rho_np1[1];
     sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
   }
 }  

 return;
}
