/************************************************************

  initialize.c

  to set up initial conditions

************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <cpml.h>
#include "basics.h"
#include "MD_FCT_defs.h"
#include "MD_FCT_utils.h"
#include "MD_FCT.h"
#include "poscol.h"
#define Std_Te 10000.0
#define Gs
/*
  Initially set up for neon.

  sigma is cross section (constant estimate) in M+2.
  m_N is neon mass in kg.

  This is only for estimating initial conditions.
*/
static double ex = 16.6, Cx = 1.5e-22;
static double ei = 21.56, Ci = .16e-21;
static double kTe, Te, Ez;
/************************************************************

 nux(x)

 excitation frequency for 3s state. Data and formula taken from Raizer.
 x is energy in eV. Thermal average used for initial conditions.

************************************************************/
double nux(x)
double x;
{
  double result;
  result = Cx*sqrt(8*x*q0/(pi*m_elec))*(ex+2*x)*exp(-ex/x);
  return(result);
}
/************************************************************

    nui(x)

      ionization frequency. Data and formula taken from Raizer.

************************************************************/
double nui(x)
double x;
{
  double result;
  result = Ci*sqrt(8*x*q0/(pi*m_elec))*(ei+2*x)*exp(-ei/x);
  return(result);
}

/************************************************************

  Initialize()

  Shottky type of distribution is assumed. Ez is taken from
  an energy balance.

  f(x) is quasi-neutral particle balance (as per Schottky) function
  for iteration procedure to determine kTe.

************************************************************/
void Initialize(set)
MD_FCT_species_set_ptr set;
{
#define f(x) (mui*x/nN- SQR(LAMBDA)*nui(x)*nN)
MD_FCT_species_ptr sp;
MD_FCT_species_set_rec_ptr rec;
static double z0 = 2.405, limit = .25, tol = .001; 
double geometric_factor, kTstd, ne0, r, z, r_fac;
double f0, eta, u_fac, A, u, ni, Eri, Gammai, E_factor;
double uk, lambda, LAMBDA, lambdaT, x, y, fx, fp, del, v, vs;
unsigned int i, j, pos, cnt, posu0, posu1;
bool success;


geometric_factor = 2*pi*j1(z0)/(z0*z0);
LAMBDA = R/z0;
kTe = k*Te;
uk = kTe/q0;
kTstd = k*Std_Temp;

mue = mue0*TORR_TO_BAR/(nN*kTstd);

x = ei/2.0;

cnt = 0; fx = 0.0;
success = false;
while(cnt < 100){
  if (cnt) {
    y = 1.001*x;
    fp = (f(y) - fx)/.001;
    del = - fx/fp;
    del = (del > limit)?limit:del;
    del = (del < -limit)?(-limit):del;
    x = x*(1.0+del);
  }
  fx = f(x);
  if (cnt&&((cnt > 100)||(success = (fabs(del) < tol)))) {break;}
  cnt++;
}
Te = (success)?(x*q0/k):Std_Te;
kTe = k*Te;
uk = kTe/q0;

lambda = .75*mue*sqrt(2.0*pi*kTe*m_elec)/q0;
lambdaT = sqrt(m_N/m_elec)*lambda;

E_factor = (mue/(mue+mui/nN))*uk/LAMBDA;
fprintf(stderr,"Te: %8.5e\n",Te);
Ez = uk*sqrt(
	     4.0/SQR(lambdaT) +
	     (mui/(nN*mue))*(3.0+ei/uk)/SQR(LAMBDA) +
	     (nux(uk)*nN/(uk*mue))*(ex/uk)
	     );

ne0 = I0/(geometric_factor*q0*mue*SQR(R)*Ez);

fprintf(stderr,"Ez: %8.5e ne0: %8.5e\n",Ez, ne0);

A = pow((m_elec/(2*pi*kTe)), 1.5);

/*
  set all other species concentrations to zero
*/
 for(rec=set->start; rec != NULL; rec = rec->next){
   sp = rec->first;
   if ((sp != etasp)&&(sp != Xrsp)&&(sp != Xzsp)&&
       (sp != Ersp)&&(sp != nNsp)&&(sp != uNsp)){
     for(i=0; i<sp->grid->alloc_size; i++){
       sp->rho0[i] = 0.0;
     }
   }
 }
/************************************************************

  First, initialize eta according to a Maxwellian in energy
  and a Bessel in radius.

  Ambipolar approximation is used for Gr (Xr). 

************************************************************/
Eri = 0.0;
for(i = 0; i <= (Nr+1); i++){
  axr->index = i;
  r = axr->r[i?axr_factor*i:axr_factor];
  z = z0*r/R;
  r_fac = (i == (Nr+1))?0.0:j0(z);
  ni = ne0*r_fac;
  nisp->rho0[i] = ni;
  Eri = (i?1.0:-1.0)*(
		     (i == (Nr+1))?
		     ((double)Nr/(double)(Nr+1))*Eri:
		     E_factor*j1(z)/r_fac
		   );
  Ersp->rho0[i] = Eri;
  Gammai = ni*mui*Eri/nN;
  Gammaisp->rho0[i] = Gammai;
  nNsp->rho0[i] = nN;
  uNsp->rho0[i] = 1.5*nN*k*Ti/q0;
  posu0 = pos = axr_factor*i;
  posu1 = posu0 + axu_factor;
  for(j = 1; j <= (Nu+1); j++){
    pos += axu_factor;
    u = axu->r[pos];
    v = STEOM*sqrt(u);
    vs = v*v;
    u_fac = A*exp(-u/uk);
    f0 = u_fac*r_fac*ne0;
    eta = v*FOURPI*f0;
/*
  now to figure out where to put it
*/
    etasp->rho0[pos] = eta;

    Xrsp->rho0[pos] = lambda*(mui/(nN*mue))*Eri/uk;
    Xzsp->rho0[pos] = -lambda*Ez/uk;
  }
  etasp->rho0[posu0] = 0.0;
  Xrsp->rho0[posu0] = Xrsp->rho0[posu1];
  Xzsp->rho0[posu0] = Xzsp->rho0[posu1];
}
Ezsp->rho0[0] = Ez;
nNsp->rho0[Nr+1] = nNsp->rho0[0] = nN;
return;
}

