/************************************************************

  poscol.c

  Simulation of DC positive column using MD_FCT
  and the elliptic representation

  Options:

  inflname take input from input file
  -i take input from initialization procedure
  -o flname    name of output file (default is "poscol.out")
  -p plot the results using PLPLOT
  -T time    run for specified amount of time (seconds)
  -I DC current
  -R tube radius
  -r Number of radial cells (Nr)
  -u Number of energy cells (Nu)

************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <cpml.h>
#include "basics.h"
#undef pi
#include "MD_FCT_defs.h"
#include "MD_FCT_utils.h"
#include "MD_FCT.h"
#include "MD_energy_axis.h"
#include "q_e_ne.h"
#include "initialize.h"
#include "sources.h"
#define pi 3.14159265
#define Std_Te 10000.0
#define VRFILE
#define Gs
/*
  Define cross sections for neon
*/

double_fn q_mom = q_e_ne_mom;
axis_ptr axu, axr, axr_config;

dblptr e_ion[] = {&ei3P2, &ei3P0, &ei_ne};
double_fn q_ion[] = {q_e_ne_3s3P2_ion, q_e_ne_3s3P0_ion, q_e_ne_ion};
string ion_targets[] = {"3s3P2", "3s3P0", "N"};
string ion_products[] = {"i", "i", "i"};

dblptr e_ex[] = {&ex3s3P1, &ex3s1P1, &ex3s3P2, &ex3s3P0, &ex3p};
double_fn q_ex[] = {q_e_ne_3s3P1, q_e_ne_3s1P1, q_e_ne_3s3P2, 
		    q_e_ne_3s3P0, q_e_ne_3p};
string ex_targets[] = {"N", "N", "N", "N", "N"};
string ex_products[] = {NULL, NULL, "3s3P2", "3s3P0", NULL};
double ex_loss_rates[] = {0.0 , 0.0 , 0.0 , 0.0 , 0.0};
/*
  Number of points, length of energy space, radius of tube,
  steady state current
*/
unsigned int Nr = 50, Nu = 100;
unsigned int Nion = 3, Nex = 5;
double R = .01, uMAX = 100.0;
double I0 = .1;
/*
  electron mobility (-torr at 300K) and "temperature" for use 
  in initialization only. Gas pressure in Torr. Gas (and ion)
  temperature in K. Ion mobility (M+2TorrV-1s-1 at 300K).
  mui0 taken from Raizer at low field....39.5 M+2 Pa V-1 s-1

  m_N is neon mass in kg.

*/
double Ti = 300.0, Ptorr = 1.0;
double mue0 = 5.0e2, mui0 = 3.0e-1;
double mui, mue;
double m_N = 3.34822e-26, qi = q0;
double m_ex = 3.34822e-26;
double nN, kappa, Cl = 3.0e-10;
double tau_z = 100e-9;         /* axial relaxation time for neutral gas */

static bool Hold_Figures = false, Append_Figures = false;
static unsigned int N_Outputs = 1;
static double Output_Time = .1e-10;

double I, Dex(double);
dblptr u = NULL, v = NULL, vs = NULL, one_v = NULL;
dblptr one_del_u = NULL, one_del_2u = NULL, one_del_2r = NULL, one_u = NULL;
dblptr one_del_r = NULL, v_2u = NULL, UTIL = NULL;
dblptr *E_ex, *E_ion, Ex_loss_rates;
dblptr ne = NULL, Gammae = NULL; 
dblptr Gammae_av = NULL, Er_av = NULL, j_ue_av = NULL;
double_fn_ptr Q_ex, Q_ion;
string *Ex_targets, *Ex_products, *Ion_targets, *Ion_products;
FILE *LOGFILE;
void cleanup(void);
MD_FCT_species_ptr etasp, Xrsp, Xzsp, nisp, Gammaisp, Ersp, Ezsp, nNsp;
MD_FCT_species_ptr uNsp, GammaNsp, tsp;
geometry_ptr configuration_space, phase_space;
unsigned int axu_factor, axr_factor;
MD_FCT_vel_list_rec_ptr eta_r_vel, eta_u_vel, Xr_r_ad_vel, Xr_u_ad_vel, Xr_r_con_vel, Xr_u_con_vel;
MD_FCT_vel_list_rec_ptr Xz_r_ad_vel, Xz_u_ad_vel, Xz_r_con_vel, Xz_u_con_vel;

double Dex(double), Di(double), get_Ti(double, double);

/************************************************************

  get_Ti(u, n)

************************************************************/
double get_Ti(u,n)
double u,n;
{
  double result;

  result = TWO_THIRDS*EOK*(u/n);
  return(result);
}
/************************************************************

   Dex(T)							     

   Diffusion coefficient for excited species in neon,
   from \cite{Phelps1959}. Units are M-1s-1.

************************************************************/
double Dex(T)
double T;
{
  return(0.76e18*pow(T,.73));
}
/************************************************************

  Di(T)

  Diffusion coefficient for ions in neon (per neon)
  = mui0*T*(1e5/760)*(1/(q0*T0))
  where mui0 is in M^2-Torr/V-s; T0 = 300K; 760mmHg/Bar

************************************************************/
double Di(T)
     double T;
{
  return(2.7378e18*mui0*T);
}
/************************************************************

  main()

************************************************************/ 
main(argc, argv)
int argc;
char **argv;
{
char **last_ptr = &(argv[argc-1]); /* pointer to last argument */
char **endptr, *chptr, nullchar;
unsigned int j, i, pos, posu1;
double x, num, denom, new_total_time, sum;
dblptr Er;
MD_FCT_species_ptr sp;
MD_FCT_species_set_ptr myset = NULL, auxset = NULL;
char flname[100], *out_file_name, *opt_ptr, *in_file_name;
bool in_flag, init_flag, make_plot = false, show = false;
bool new_total_time_flag = false;
dblptr t1 = NULL;

atexit(cleanup);
init_flag = in_flag = false;
in_file_name = "poscol.in";
out_file_name = "poscol.out";
MD_trace = false;
MD_DATADIR = "/home/richley/simulations/poscol/X/data";
/*
  process options; get output file name, if any. 
  Default output is "poscol.out"
  -i: take initial data from simplified model
  -o flname: direct output to flname
  -p: generate MATLAB plots
  -t: set MD_trace = true
  -T: simulation time
  -d: set MD_DATADIR to string
  -a: set MD_Total_time to value
*/
chptr = &nullchar;
endptr = &chptr;
 while(++argv <= last_ptr){
  if((opt_ptr = *argv)[0] == '-'){
    for(i=1; opt_ptr[i] != '\0'; i++){
      switch(opt_ptr[i])
	{
	case 'o':
	  if (argv <= last_ptr) {
	    out_file_name = *(++argv);
	  }
	  break;
	case 'd':
	  if (argv <= last_ptr) {
	    MD_DATADIR = *(++argv);
	  }
	  break;
	case 'i':
	  init_flag = true;
	  break;
	case 'p':
	  make_plot = true;
	  break;
	case 's':
	  show = true;
	  break;
	case 't': 
	  MD_trace = true;
	  break;
	case 'I':
	  if (argv <= last_ptr){
	    I0 = strtod(*(++argv),endptr);
	  }
	  if (**endptr != '\0'){
	    fprintf(stderr,"poscol: bad current: %s \n",*argv);
	    exit(1);
	  }
	  break;
	case 'r':
	  if (argv <= last_ptr){
	    Nr = (unsigned int)strtoul(*(++argv),endptr,10);
	  }
	  if (**endptr != '\0'){
	    fprintf(stderr,"poscol: bad Nr: %s \n",*argv);
	    exit(1);
	  }
	  break;
	case 'u':
	  if (argv <= last_ptr){
	    Nu = (unsigned int)strtoul(*(++argv),endptr,10);
	  }
	  if (**endptr != '\0'){
	    fprintf(stderr,"poscol: bad Nu: %s \n",*argv);
	    exit(1);
	  }
	  break;
	case 'R':
	  if (argv <= last_ptr){
	    R = strtod(*(++argv),endptr);
	  }
	  if (**endptr != '\0'){
	    fprintf(stderr,"poscol: bad radius: %s \n",*argv);
	    exit(1);
	  }
	  break;
	case 'a':
	  if (argv <= last_ptr){
	    new_total_time = strtod(*(++argv),endptr);
	  }
	  if (**endptr != '\0'){
	    fprintf(stderr,"poscol: bad time value: %s \n",*argv);
	    exit(1);
	  }
	  new_total_time_flag = true;
	  break;
	case 'T':
	  if (argv <= last_ptr){
	    Output_Time = strtod(*(++argv),endptr);
	  }
	  if (**endptr != '\0') {
	    fprintf(stderr,"poscol: bad Output Time: %s\n",*argv);
	    exit(1);
	  }
	  break;
	default:
	  fprintf(stderr,"poscol: bad option %c\n",(*argv)[i]);
	  exit(1);
	  break;
	}
      if (opt_ptr != *argv) {break;}
    }
  } else {
    if (!in_flag){
      in_file_name = *argv;
      in_flag = true;

    } else {
      fprintf(stderr,"poscol: unknown parameter %s\n",(*argv));
    }
  }
}
Q_ex = &q_ex[0]; Q_ion = &q_ion[0];
E_ex = &e_ex[0]; E_ion = &e_ion[0];
Ex_targets = &ex_targets[0]; Ex_products = &ex_products[0];
Ion_targets = &ion_targets[0]; Ion_products = &ion_products[0];
Ex_loss_rates = ex_loss_rates;
/* 
  Create new species set "MySet" and grid "MyGrid"
*/
myset = MD_FCT_new_species_set("MySet");
configuration_space = MD_FCT_new_grid("Configuration_Space");
phase_space = MD_FCT_new_grid("Phase_Space");
/*
  Define axes for Radial as defined by "MD_FCT_cyl"
  of length R with Nr points. Define kinetic energy
  axis u of length Nu. 

  Get various vectors with energy and velocity data.
*/
 axr_config = MD_FCT_new_axis("r", "M", Nr, configuration_space, R, MD_FCT_cyl);
 axr = MD_FCT_new_axis("r", "M", Nr, phase_space, R, MD_FCT_cyl);
 axu = MD_FCT_new_axis("u", "eV", Nu, phase_space, uMAX, MD_energy_axis);
 axr_factor = axr->mul;
 axu_factor = axu->mul;
 u = get_memory(u, (Nu+2), double);
 one_u = get_memory(one_u, (Nu+2), double);
 one_v = get_memory(one_v, (Nu+2), double);
 v = get_memory(v, (Nu+2), double);
 vs = get_memory(vs, (Nu+2), double);
 t1 = get_memory(t1, (Nu+2), double);
 one_del_u = get_memory(one_del_u, (Nu+2), double);
 one_del_2u = get_memory(one_del_2u, (Nu+2), double);
 one_del_r = get_memory(one_del_r, (Nr+2), double);
 one_del_2r = get_memory(one_del_2r, (Nr+2), double);
 v_2u = get_memory(v_2u, (Nu+2), double);
 UTIL = get_memory(UTIL, (Nr+2)*(Nu+2), double);
 for(j=1; j<=(Nu+1); j++){
   u[j] = axu->r[j*axu->mul];
   one_u[j] = 1.0/u[j];
   vs[j] = TEOM*u[j];
   v[j] = STEOM*sqrt(u[j]);
   v_2u[j] = .5*v[j]/u[j];
   one_v[j] = 1.0/v[j];
 }
 u[0] = axu->r[0];
 one_u[0] = one_u[1];
 v[0] = -v[1];
 one_v[0] = 1.0/v[0];
 v_2u[0] = -.5*v[0]/u[0];
 for(j=0; j<=Nu; j++){
   one_del_u[j] = 1.0/(u[j+1]-u[j]);
 }
 for(j=1; j<=Nu; j++){
   one_del_2u[j] = 1.0/(u[j+1]-u[j-1]);
 }
 for(i=0; i<=Nr; i++){
   one_del_r[i] = 1.0/(axr_config->r[i+1]-axr_config->r[i]);
 }
 for(i=1; i<=Nr; i++){
   one_del_2r[i] = 1.0/(axr->r[(i+1)*axr->mul] - axr->r[(i-1)*axr->mul]);
 }
/*
  Define all the species, and set boundary conditions
*/
 etasp = sp = MD_FCT_new_species("eta", "s+2/M+5", myset, phase_space);
 eta_u_vel = MD_FCT_create_velocity(etasp, "u", geometric_conservation, "u-velocity", "eV/s");
 eta_r_vel = MD_FCT_create_velocity(etasp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc_fn(sp, "r", right, local, eta_wall);
 MD_FCT_set_bc(sp, "u", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc_fn(sp, "u", right, local, eta_OutFlow_u);
 etasp->Flux_Limiter = strong;
 etasp->enable_minimum = TRUE;
 etasp->minimum = 0.0;

 Xrsp = sp = MD_FCT_new_species("Xr", "", myset, phase_space);
 Xr_u_ad_vel = MD_FCT_create_velocity(Xrsp, "u", advection, "u-advection_velocity", "eV/s");
 Xr_r_ad_vel = MD_FCT_create_velocity(Xrsp, "r", advection, "r-advection_velocity", "M/s");
 Xr_u_con_vel = MD_FCT_create_velocity(Xrsp, "u", conservation, "u-conservation_velocity", "eV/s");
 Xr_r_con_vel = MD_FCT_create_velocity(Xrsp, "r", conservation, "r-conservation_velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "u", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "u", right, SYMMETRIC, ZERO, local);
 Xrsp->Flux_Limiter = strong;
 Xrsp->enable_minimum = TRUE;
 Xrsp->enable_maximum = TRUE;
 Xrsp->minimum = -1.0;
 Xrsp->maximum = 1.0;

 Xzsp = sp = MD_FCT_new_species("Xz", "", myset, phase_space);
 Xz_u_ad_vel = MD_FCT_create_velocity(Xzsp, "u", advection, "u-advection_velocity", "eV/s");
 Xz_r_ad_vel = MD_FCT_create_velocity(Xzsp, "r", advection, "r-advection_velocity", "M/s");
 Xz_u_con_vel = MD_FCT_create_velocity(Xzsp, "u", conservation, "u-conservation_velocity", "eV/s");
 Xz_r_con_vel = MD_FCT_create_velocity(Xzsp, "r", conservation, "r-conservation_velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "u", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "u", right, SYMMETRIC, ZERO, local);
 Xzsp->Flux_Limiter = strong;
 Xzsp->enable_minimum = TRUE;
 Xzsp->enable_maximum = TRUE;
 Xzsp->minimum = -1.0;
 Xzsp->maximum = 1.0;

 nisp = sp = MD_FCT_new_species("ni", "M-3", myset, configuration_space);
 MD_FCT_create_velocity(nisp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 sp->charge = +q0;
 sp->mass = m_N;
 sp->Flux_Limiter = strong;
 sp->D[0] = get_memory(sp->D[0], sp->grid->alloc_size, double);
 for(i=0; i<sp->grid->alloc_size; i++){
   sp->D[0][i] = Di(Ti);
 }

 Gammaisp = sp = MD_FCT_new_species("Gammai", "M-2s-1", myset, configuration_space);
 MD_FCT_create_velocity(Gammaisp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 Gammaisp->Flux_Limiter = strong;

 sp = MD_FCT_new_species("n3s3P2", "M-3", myset, configuration_space);
 MD_FCT_create_velocity(sp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 sp->charge = 0.0;
 sp->mass = m_ex;
 sp->Flux_Limiter = strong;
 sp->D[0] = get_memory(sp->D[0], sp->grid->alloc_size, double);
 for(i=0; i<sp->grid->alloc_size; i++){
   sp->D[0][i] = Dex(Ti);
 }

 sp = MD_FCT_new_species("Gamma3s3P2", "M-2s-1", myset, configuration_space);
 MD_FCT_create_velocity(sp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 sp->Flux_Limiter = strong;

 sp = MD_FCT_new_species("n3s3P0", "M-3", myset, configuration_space);
 MD_FCT_create_velocity(sp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 sp->charge = 0.0;
 sp->mass = m_ex;
 sp->Flux_Limiter = strong;
 sp->D[0] = get_memory(sp->D[0], sp->grid->alloc_size, double);
 for(i=0; i<sp->grid->alloc_size; i++){
   sp->D[0][i] = Dex(Ti);
 }

 sp = MD_FCT_new_species("Gamma3s3P0", "M-2s-1", myset, configuration_space);
 MD_FCT_create_velocity(sp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, 1.0, ZERO, local);
 sp->Flux_Limiter = strong;
 
 Ersp = sp = MD_FCT_new_species("Er", "V/M", myset, configuration_space);
 MD_FCT_create_velocity(Ersp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc_fn(sp, "r", right, local, Er_wall);
 Ersp->Flux_Limiter = strong;

 nNsp = sp = MD_FCT_new_species("nN", "M-3", myset, configuration_space);
 MD_FCT_create_velocity(nNsp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, SYMMETRIC, ZERO, local);
 sp->mass = m_N;
 sp->charge = 0.0;
 sp->Flux_Limiter = strong;

 GammaNsp = sp = MD_FCT_new_species("GammaN", "M-2s-1", myset, configuration_space);
 MD_FCT_create_velocity(GammaNsp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, ANTI_SYMMETRIC, ZERO, local);
 MD_FCT_set_bc(sp, "r", right, ANTI_SYMMETRIC, ZERO, local);
 sp->charge = 0.0;
 sp->Flux_Limiter = strong;

 uNsp = sp = MD_FCT_new_species("uN", "eVM-3", myset, configuration_space);
 MD_FCT_create_velocity(uNsp, "r", conservation, "r-velocity", "M/s");
 MD_FCT_set_bc(sp, "r", left, SYMMETRIC, ZERO, local);
 MD_FCT_set_bc_fn(sp, "r", right, local, uN_wall);
 sp->charge = 0.0;
 sp->Flux_Limiter = strong;

 Ezsp = sp = MD_FCT_new_species("Ez", "V/M", myset, DUMMY_GRID);

 fprintf(stderr,"\n");
 strcpy(flname, myset->name);strcat(flname,".log");
 MD_LOGFILE = LOGFILE = fopen(flname, "w");

 ne = get_memory(ne, (Nr+2), double);
 Gammae = get_memory(Gammae, (Nr+2), double);
 Gammae_av = get_memory(Gammae_av, (Nr+2), double);
 Er_av = get_memory(Er_av, (Nr+2), double);
 j_ue_av = get_memory(j_ue_av, (Nr+2), double);

/*
  Get some important parameters, and
  Initialize the concentration data
*/
 nN = Ptorr*TORR_TO_BAR/(k*Ti);
 mui = mui0*TORR_TO_BAR/(k*Std_Temp);
 fprintf(stdout,"poscol: running with DC current: %8.5e A radius: %8.5e M\n",I0, R);
 fprintf(stdout,"poscol: %u radial cells %u energy cells\n",Nr, Nu);

 if (init_flag) {
   fprintf(stderr,"poscol: initializing from approximate model\n");
   Initialize(myset);
 } else {
   fprintf(stderr,"poscol: taking input from %s\n",in_file_name);
   MD_FCT_INPUT_FILE = fopen(in_file_name, "r");
   if (MD_FCT_INPUT_FILE == NULL){
     fprintf(stderr,"poscol: unable to find input file %s\n",in_file_name);
     exit(1);
   }
   MD_FCT_read_all_rho(myset);
   fclose(MD_FCT_INPUT_FILE);
 }
 if (new_total_time_flag){
   MD_Total_time = new_total_time;
 }
/*
  initialize ne, Gammae, Gammae_av, Er_av
*/
 for(i=1; i<=Nr; i++){
   axr->index = i;
   posu1 = i*axr->mul + axu->mul;
   ne[i] = EOM*MD_integral_dVol(etasp->rho0, axu, phase_space);
   for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
     t1[j] = Xrsp->rho0[pos]*etasp->rho0[pos]*sqrt(axu->r[pos]);
   }
   Gammae[i] = STEOM*EOM*MD_integral_dV(t1,axu, phase_space);  
   Gammae_av[i] = Gammae[i];
   Er_av[i] = Ersp->rho0[i];
 }
 ne[0] = ne[1]; Gammae[0] = -Gammae[1];
 ne[Nr+1] = 0.0; Gammae[Nr+1] = Gammae[Nr];
 Gammae_av[0] = -Gammae_av[1]; Gammae_av[Nr+1] = Gammae_av[Nr];
 Er_av[0] = -Er_av[0]; Er_av[Nr+1] = Er_av[Nr];

 fprintf(stderr,"poscol: directing output to %s\n", out_file_name);
 MD_FCT_OUTPUT_FILE = fopen(out_file_name, "w");
 if (MD_FCT_OUTPUT_FILE == NULL){
   fprintf(stderr,"poscol: unable to open output file %s\n",out_file_name);
   exit(1);
 }
/*
  Flow MySet for time Output_Time, with velocities and sources
  defined by procedure "sources()" , and plot the results
*/
 if (Output_Time < 0.0) {
   fprintf(stderr,"Error: negative Output Time: %8.5e\n",Output_Time);
   exit(1);
 }
 fprintf(stderr,"simulating for: %8.5e sec\n",Output_Time);

 auxset = MD_FCT_new_species_set("AuxSet");

 for(j=0; j<N_Outputs; j++){
   MD_Flow(myset, sources, Output_Time);
   MD_FCT_print_all_rho(myset);
 }


 sp = MD_FCT_new_species("I", "A", auxset, DUMMY_GRID);
 sp->rho0[0] = I;

  sp = MD_FCT_new_species("G_r", "s+2/M+5", auxset, phase_space);
  for(i=0; i<=Nr+1; i++){
    for(j=0; j<=Nu+1; j++){
      pos = i*axr->mul + j*axu->mul;
      sp->rho0[pos] = etasp->rho0[pos]*Xrsp->rho0[pos];
    }
  }
  sp->ViewAngle = "[30 70]";

  sp = MD_FCT_new_species("G_z", "s+2/M+5", auxset, phase_space);
  for(i=0; i<=Nr+1; i++){
    for(j=0; j<=Nu+1; j++){
      pos = i*axr->mul + j*axu->mul;
      sp->rho0[pos] = etasp->rho0[pos]*Xzsp->rho0[pos];
    }
  }
  sp->ViewAngle = "[30 50]";

  sp = MD_FCT_new_species("Xs", "", auxset, phase_space);
  for(i=0; i<=Nr+1; i++){
    for(j=1; j<=Nu+1; j++){
      pos = i*axr->mul + j*axu->mul;
      num = SQR(Xrsp->rho0[pos])+SQR(Xzsp->rho0[pos]);
      sp->rho0[pos] = num;
    }
  }
  sp->ViewAngle = "[30 50]";

  sp = MD_FCT_new_species("n", "M-6 s+3", auxset, phase_space);
  for(i=0; i<=Nr+1; i++){
    for(j=1; j<=Nu+1; j++){
      pos = i*axr->mul + j*axu->mul;
      sp->rho0[pos] = etasp->rho0[pos]/v[j];
    }
    sp->rho0[i*axr->mul] = sp->rho0[i*axr->mul + axu->mul];
  }
  sp->ViewAngle = "[30 50]";

  sp = MD_FCT_new_species("vi", "M/s", auxset, configuration_space);
  sp->rho0 = nisp->vel->velocity;

  sp = MD_FCT_new_species("ne", "M-3", auxset, configuration_space);
  sp->rho0 = ne;

  sp = MD_FCT_new_species("Gammae", "M-2s-1", auxset, configuration_space);
  /*  sp->rho0 = Gammae;*/
  for(axr->index=0; axr->index<=Nr; axr->index++){
    posu1 = axr->index*axr_factor + axu_factor;
    sum = 0.0;
    for(axu->index = 1, pos = posu1; axu->index <= Nu; axu->index++, pos += axu_factor){
      sum += .25*(etasp->rho0[pos]+etasp->rho0[pos+axr_factor])*
	(eta_r_vel->velocity[pos]+eta_r_vel->velocity[pos+axr_factor])*axu->Volume[pos];
    }
    sp->rho0[axr->index] = EOM*sum + ((axr->index == Nr)?
				      ((-electron_wall_rate*axr_config->Volume[Nr])/(2*pi*R))
				      :0.0);
  }

  sp = MD_FCT_new_species("jue", "eV/(s M+2)", auxset, configuration_space);

  for(axr->index=1; axr->index<=Nr+1; axr->index++){
    posu1 = axr->index*axr_factor + axu_factor;
    for(axu->index = 1, pos = posu1; axu->index <= Nu; axu->index++, pos += axu_factor){
      t1[axu->index] = axu->r[pos]*Xrsp->rho0[pos]*etasp->rho0[pos]*sqrt(axu->r[pos]);
    }
    sp->rho0[axr->index] = STEOM*EOM*MD_integral_dV(t1,axu, phase_space);  
  }
  sp->rho0[0] = -sp->rho0[1];

  sp = MD_FCT_new_species("u_av", "eV", auxset, configuration_space);
  for(i=1; i<=Nr; i++){
    axr->index = i;
    posu1 = i*axr_factor+axu_factor;
    for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
      t1[j] = etasp->rho0[pos]*axu->r[pos];
    }
    sp->rho0[i] = MD_integral_dV(t1, axu, phase_space)/
      MD_integral_dVol(etasp->rho0, axu, phase_space);
  }
  sp->rho0[0] = sp->rho0[1];  sp->rho0[Nr+1] = 0.0;

  sp = MD_FCT_new_species("Ti", "K", auxset, configuration_space);

  for(i=1; i<=Nr+1; i++){
    sp->rho0[i] = get_Ti(uNsp->rho0[i],nNsp->rho0[i]);
  }
  sp->rho0[0] = sp->rho0[1];

  sp = MD_FCT_new_species("Potential", "V", auxset, configuration_space);
  Er = Ersp->rho0;
  for(i=1; i<=Nr; i++){
    if (i == 1) {
      sp->rho0[0] = sp->rho0[1] = -.5*Er[1]*axr_config->r[1];
    } else {
      sp->rho0[i] = sp->rho0[i-1] - .5*(Er[i]+Er[i-1])*(axr_config->r[i]-axr_config->r[i-1]);
    }
  }
  sp->rho0[Nr+1] = sp->rho0[Nr] - .5*(Er[Nr+1]+Er[Nr])*(axr_config->r[Nr+1]-axr_config->r[Nr]);

  sp = MD_FCT_new_species("e_Collisional_exch", "Js-1M-3", auxset, configuration_space);
  sp->rho0 = exch;
  sp->rho0[0] = sp->rho0[1];
  sp->rho0[Nr+1] = 0.0;

  sp = MD_FCT_new_species("Compression", "Js-1M-3", auxset, configuration_space);
  sp->rho0 = comp;
  sp->rho0[0] = sp->rho0[1];
  sp->rho0[Nr+1] = 0.0;

  sp = MD_FCT_new_species("Conduction", "Js-1M-3", auxset, configuration_space);
  sp->rho0 = cond;
  sp->rho0[0] = sp->rho0[1];
  sp->rho0[Nr+1] = 0.0;

  sp = MD_FCT_new_species("Gammae_av", "M-2 s-2", auxset, configuration_space);
  sp->rho0 = Gammae_av;

  sp = MD_FCT_new_species("Er_av", "V/M", auxset, configuration_space);
  sp->rho0 = Er_av;

  fprintf(stdout,"Ion_Wall_Rate: %8.5e\n",i_Wall_Rate);

 MD_output_all_rho(myset);
 MD_output_all_rho(auxset);
 MD_output_all_vel(etasp);
 MD_output_all_vel(Xrsp);
 MD_output_all_vel(Xzsp);
 MD_output_all_vel(nNsp);
fclose(MD_FCT_OUTPUT_FILE);
fclose(LOGFILE);
return(0);
}

/*
  exit function
*/
void cleanup(){
  MD_FCT_species_set_ptr tmpset = NULL;
  MD_FCT_species_ptr sp;
#if 0
  MD_DATADIR = "/home/richley/simulations/poscol/X/data";
  tmpset = MD_FCT_new_species_set("TmpSet");

  sp = MD_FCT_new_species("eta_incr", "M-3/s", tmpset, phase_space);
  sp->rho0 = UTIL;
  sp->ViewAngle = "[30 70]";

  MD_output_all_rho(tmpset);
#endif
}
