/************************************************************

  sources.c

  source terms for poscol simulation

************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#define ccc
#ifdef ccc
#include <cpml.h>
#endif
#include "basics.h"
#undef pi
#include "MD_FCT_defs.h"
#include "MD_FCT.h"
#include "MD_FCT_utils.h"
#include "poscol.h"
#include "special.h"
#include "collisions.h"
#define pi 3.14159265
#define POSCOL_Xs_LIMIT 1.0
#define TAU_AV 10.0e-9

static MD_FCT_species_ptr *exsp = NULL;
unsigned int rusize, u_infty, r_R; 
dblptr del_elastic = NULL, ex_W = NULL, ionization = NULL;
dblptr u_diff = NULL;
dblptr eta_TBC = NULL;
double delel, kTi_e, kTi_M, Vx, qi_M, ion_a1, ion_a2;
double ion_wall_rate, electron_wall_rate, i_Wall_Rate;
double dtkinetic;

static dblptr t1 = NULL, t2 = NULL, Ps = NULL, Ve = NULL, C = NULL;
static dblptr Vs = NULL, Xterm = NULL, log_n = NULL; 
dblptr exch = NULL, comp = NULL, cond = NULL;
static dblptr kTi_ = NULL;
static double ion_elastic;
static double Er_BC_S, Er_BC_V, kTi, one_u_amb;
static unsigned int Nexp = 0;
static double ni_W, Gammai_W;
double mindu_v, ion_sound_speed, vchar, Lchar;
static unsigned int Xscnt;

void get_heavy_wall_terms(MD_FCT_species_ptr, dblptr, MD_FCT_species_ptr, dblptr);
bool sources(MD_FCT_species_set_ptr, double);
void eta_wall(MD_FCT_bc *);
void ni_wall(MD_FCT_bc *);
void Er_wall(MD_FCT_bc *);
void eta_OutFlow_u(MD_FCT_bc *);
void Gammai_wall(MD_FCT_bc *);

static bool advance_sources(MD_FCT_species_set_ptr, double);

/************************************************************

  eta_OutFlow_u()

  extrapolarion boundary condition

************************************************************/
void eta_OutFlow_u(bc)
MD_FCT_bc *bc;
{
  bc->vchar_r = 0.0;
  bc->Lchar_r = 0.0;
  bc->Sr = eta_TBC[axr->index]; 
  bc->Vr = 0.0;
  bc->rho_amb_r = 0.0;
}
/************************************************************

  eta_wall()

  wall boundary condition for eta

************************************************************/
void eta_wall(bc)
MD_FCT_bc *bc;
{

  bc->Sr = SBC[axu->index];
  bc->Vr = 0.0;

  bc->Lchar_r = 0.0;
}
/************************************************************

  uN_wall()

  wall boundary condition for uN

  Keeps 1.5 kT/e at wall equal to 1.5kTi/e

************************************************************/
void uN_wall(bc)
MD_FCT_bc *bc;
{
  bc->Sr = ANTI_SYMMETRIC;
  bc->Vr = 3.0*kTi_e*nNsp->rho0[Nr];
  bc->Lchar_r = 0.0;
}
/************************************************************

  Er_wall()

  wall boundary condition for Er

************************************************************/
void Er_wall(bc)
MD_FCT_bc *bc;
{
  bc->Sr = Er_BC_S;
  bc->Vr = Er_BC_V;
  bc->Lchar_r = 0.0;
}
/************************************************************

  get_heavy_wall_terms()

  gets ni_W, Gammai_W, ni_BC, Gammai_BC for
  any heavy species

************************************************************/
void get_heavy_wall_terms(sp, n_W, Gammasp, Gamma_W)
MD_FCT_species_ptr sp, Gammasp;
dblptr n_W, Gamma_W;
{
  double d, x, c1, c2, E4, E4_, cs, x1, x2;
  double sound_speed, vr, n_BC, ion_elastic_, result;

  ion_elastic_ = ion_elastic*nNsp->rho0[Nr];
  vr = sp->vel->velocity[Nr];
  sound_speed = sqrt((2.0*k*Ti)/sp->mass);
  d = (R - axr_config->r[Nr]);
  c1 = (vr+SQRT__PI_2*sound_speed);
  x1 = (d*ion_elastic_)/c1;
  E4 = (c1>0.0)?En(x1,4):0.0;
  c2 = (vr-SQRT__PI_2*sound_speed);
  x2 = (d*ion_elastic_)/c2;
  E4_ = (c2>0.0)?En(x2,4):0.0;
  n_BC = 6.0*(E4 + E4_) - 1.0;
  sp->bc[axr_config->sequence].Sr = n_BC;
  Gammasp->bc[axr_config->sequence].Sr = n_BC;
  cs = sound_speed*sound_speed;
  x = (E4 - E4_)/(sqrt_2*SQRT_PI*d*sound_speed);
  *n_W = 1.5*(vr*vr+PI_2*cs)*x;
  *Gamma_W = (vr*vr + 1.5*PI*cs)*x;
  return;
}
/************************************************************

  kappaN()

  Thermal conductivity for neutral gas
  From Chapman and Cowling, scaled according to Vincenti
  and Kruger units are eV J-1 M-1 s-1

************************************************************/
static double kappaN(kT)
double kT;
{
  double result;

  result = 3.42313e50*sqrt(kT);
  return(result);
}

/************************************************************

  advance_sources()

************************************************************/
bool advance_sources(set, dt)
double dt;
MD_FCT_species_set_ptr set;
{
MD_FCT_species_ptr sp, Gsp;
register dblptr eta_n, eta_np1, Xr_n, Xr_np1, Xz_n, Xz_np1;
dblptr ni_n, ni_np1, Gammai_n,  Gammai_np1;
dblptr Ez_n, Ez_np1, Er_n, Er_np1;
dblptr nN_n, nN_np1, GammaN_n, GammaN_np1, uN_n, uN_np1;
dblptr r_Area, r_Volume, Vs_, Ps_ ;
unsigned int i, j, next_pos, pos, posu0, posu1, posui;
double dtq_eps0, dtkTi_M, dtqi_M, incr;
double Xr_, Xz_, eta_, vj, Xterm_, Psp;
double one_coll_G, one_coll_Gammai, one_coll_ni;
double G_eff, one_coll_ex, dtom, d1, Er_, Ez_, EdotX;
double x,  Xs, ddr_log_eta, ddu_log_eta, f1, f2, v1, v2, dd_2r;
double Adt, Xmag, denom, term1, term2, E, Eterm, vterm, rterm;
double GammaN, uN, dt_M, nN_, kTi__, dtomega, dtomegai, psum, psum_; 
double ddu_log_n, div_vz, ri;
double wall_rate, flux, e_wall_rate, i_wall_rate, drN;

 drN = .5*(
	SQR(axr_config->r[Nr]) - SQR(.5*(axr_config->r[Nr] + axr_config->r[Nr-1]))
	)/axr_config->r[Nr];
 Ps_ = Ps; Vs_ = Vs; 
 r_Area = axr_config->Area;
 r_Volume = axr_config->Volume;
 eta_n = etasp->rho_n;
 eta_np1 = etasp->rho_np1;
 Xr_n = Xrsp->rho_n;
 Xr_np1 = Xrsp->rho_np1;
 Xz_n = Xzsp->rho_n;
 Xz_np1 = Xzsp->rho_np1;
 ni_n = nisp->rho_n;
 ni_np1 = nisp->rho_np1;
 Gammai_n = Gammaisp->rho_n;
 Gammai_np1 = Gammaisp->rho_np1;
 Er_n = Ersp->rho_n;
 Er_np1 = Ersp->rho_np1;
 Ez_n = Ezsp->rho_n;
 Ez_np1 = Ezsp->rho_np1;
 nN_n = nNsp->rho_n;
 nN_np1 = nNsp->rho_np1;
 GammaN_n = GammaNsp->rho_n;
 GammaN_np1 = GammaNsp->rho_np1;
 uN_n = uNsp->rho_n;
 uN_np1 = uNsp->rho_np1;

 dtq_eps0 = dt*q0/eps0;
 dtkTi_M = dt*kTi_M; 
 dtqi_M = dt*qi_M;
 dt_M = dt/m_N;
/************************************************************
  
  Get diffusion solution for finite Ti...take eta_n -> eta_np1

  might be a mistake here....might be missing uN in t1

************************************************************/
 for(i=1; i<=Nr; i++){
   posu0 = i*axr_factor; 
   uN = uN_n[i];
   nN_ = nN_n[i];
   for(j=0, pos = posu0; j<=(Nu); j++, pos = next_pos){
     next_pos = pos + axu_factor;
     t1[j] = u_diff[j]*(eta_n[next_pos]-eta_n[pos])*one_del_u[j];
   }
   posu1 = posu0 + axu_factor;   
   for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
     incr = nN_*2.0*(t1[j]-t1[j-1])*one_del_2u[j];
     t2[j] = incr*u[j] - del_elastic[j]*(nN_*u[j] - uN)*eta_n[pos];
     eta_ = eta_np1[pos] = eta_n[pos] + incr*dt;
     Xterm[pos] = incr/eta_;
   }
   eta_np1[pos] = eta_TBC[i]*eta_np1[pos-axu_factor];
   eta_np1[posu0] = etasp->bc[axu->index].Sl*eta_np1[posu1];
   exch[i] = EOM*MD_integral_dV(t2, axu, phase_space);
   kTi_[i] = TWO_THIRDS*q0*uN_n[i]/nN_n[i];
 }

 kTi_[0] = kTi_[1];
 kTi_[Nr+1] = 2.0*kTi - kTi_[Nr];
/************************************************************

  put together the inelastic transport matrix
  (1 + de_excitation_dat*dt)/(1 + inelastic*dt)
  and solve it. Take eta_np1 -> eta_np1.

************************************************************/
 Apply_inelastics(eta_np1, dt, true, true, Xterm, nN_n, uN_n);
 posu0 = 0;
 for(j=0, pos = posu0; j<=(Nu); j++, pos += axu_factor){
   eta_np1[pos] = eta_np1[pos+axr_factor];
 }
 posu0 = (Nr+1)*axr_factor;
 for(j=0, pos = posu0; j<=(Nu); j++, pos += axu_factor){
   eta_np1[pos] = SBC[j]*eta_np1[pos-axr_factor];
 }
/************************************************************
  Apply the wall absorption terms and boundary conditions
************************************************************/

 electron_wall_rate = do_wall_absorption(Xr_n, Xz_n, eta_np1, dt);

 /************************************************************

  Energy source terms for neutrals

 ************************************************************/
 for(i=0; i<=Nr; i++){
   kTi__ = .5*(kTi_[i]+kTi_[i+1]);
   C[i] = kappaN(kTi__)*(kTi_[i+1]-kTi_[i])*one_del_r[i];
 }
 for(i=1; i<=Nr; i++){
   f1 = r_Area[i]/r_Volume[i];
   f2 = r_Area[i-1]/r_Volume[i];
   v1 = .5*(GammaN_n[i+1]/nN_n[i+1] + GammaN_n[i]/nN_n[i]);
   v2 = .5*(GammaN_n[i-1]/nN_n[i-1] + GammaN_n[i]/nN_n[i]);
   /* do axial relaxation on neutrals */
   nN_np1[i] = nN_n[i] - dt*(nN_n[i]*(div_vz = - (1 - uN_n[i]*one_u_amb)/tau_z));
   uN_np1[i] = uN_n[i] + dt*
     (
      (cond[i] = (f1*C[i] - f2*C[i-1])) - 
      (comp[i] = TWO_THIRDS*uN_n[i]*(f1*v1-f2*v2 + div_vz)) - exch[i]
      );
 }
 uN_np1[0] = uN_np1[1];
 uN_np1[Nr+1] = 3.0*kTi_e*nNsp->rho0[Nr] - uN_np1[Nr];
 nN_np1[0] = nN_np1[1]; nN_np1[Nr+1] = nN_np1[Nr];
/************************************************************

  Get log_n

************************************************************/
 for(i=0; i<=(Nr+1); i++){
   posu0 = i*axr_factor;
   posu1 = posu0+axu_factor;
   for(j=1, pos = posu1; j<=(Nu+1); j++, pos += axu_factor){
     x = eta_np1[pos]*one_v[j];
     log_n[pos] = (x>0.0)?log(x):0.0;
   }
   log_n[posu0] = log_n[posu1];
 }

/************************************************************

  Put together phase space pressure and drift arrays

  Also, do rotation of X

************************************************************/
 Ez_ = *Ez_n; 
 for(i=0; i<=Nr+1; i++){
   posu0 = i*axr_factor;
   Er_ = Er_n[i]; 
   E = sqrt(Er_*Er_+Ez_*Ez_);
   posu1 = posu0 + axu_factor;
   for(j=0, pos = posu0; j<=Nu+1; j++, pos += axu_factor){
     Xs = SQR(Xr_n[pos])+SQR(Xz_n[pos]);
     vj = v[j];
     Xmag = sqrt(Xs);
     Ps_[pos] = get_Pterm(Xs);
     Vs_[pos] = vj*((vterm = get_Vterm(Xs)) - 1.0);
     if ((i >= 1) && (i <= Nr) && (j >= 1) && (j <= Nu)){
       EdotX = Er_*Xr_n[pos]+Ez_*Xz_n[pos];
       Adt = Xmag*vterm*v_2u[j]*E*dt;
       x = 1.0/(1.0+Adt);
       denom = E*(Xmag*E*(1+x*x)+EdotX*(x*x-1));
       term1 = Xs*E*(x*x-1)+Xmag*EdotX*SQR((x-1));
       term2 = 2*Xmag*E*E*x;
       Xr_np1[pos] = (term1*Er_ + term2*Xr_n[pos])/denom;
       Xz_np1[pos] = (term1*Ez_ + term2*Xz_n[pos])/denom;
     }
   }
   posui = posu1 + Nu*axu_factor;
   Xr_np1[posu0] = Xr_np1[posu1];
   Xz_np1[posu0] = Xz_np1[posu1];
   Xr_np1[posui] = Xr_np1[posui-axu_factor];
   Xz_np1[posui] = Xz_np1[posui-axu_factor];
 } 
/*
  boundary terms for ions and excited species
*/
 one_coll_ni = 1.0/(1.0+dt*ni_W);
 ni_np1[Nr] *= one_coll_ni;
 ion_wall_rate = -ni_np1[Nr]*ni_W;
 ni_np1[Nr+1] = nisp->bc[axr_config->sequence].Sr*ni_np1[Nr];

 for(j=0; j<2*Nexp; j+=2){
   sp = exsp[j];
   Gsp = exsp[j+1];
   sp->rho_np1[Nr] /= (1.0 + dt*ex_W[j]);
   sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
 }
/************************************************************

  get parameters for ion wall collision rate
  get new ion concentrations from source terms

  get C, ne, Gammae
  
  get I

************************************************************/
 for(i=1; i<=Nr+1; i++){
   posu1 = i*axr_factor + axu_factor;
   axr_config->index = (axr->index = i);
   for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
     vj = v[j]; eta_ = eta_np1[pos];
     t1[j] = Xr_n[pos]*vj*eta_;
     t2[j] = Xz_n[pos]*vj*eta_;
   }
   ne[i] = EOM*MD_integral_dVol(eta_np1, axu, phase_space);
   Gammae[axr->index] = EOM*MD_integral_dV(t1,axu, phase_space);  
   C[axr->index] = MD_integral_dV(t2, axu, phase_space);
 }
 Gammae[0] = -Gammae[1];
 ne[0] = ne[1];
 C[0] = C[1];
 I = -EOM*q0*MD_integral_dVol(C, axr_config, configuration_space);
 (*Ez_np1) = ((*Ez_n) + dt*(I0 - I)/Cl);

/************************************************************

  get Er, Ez, Gammai, Gammaex

************************************************************/
 Ez_ = *Ez_np1; 
 for(i=1; i<=Nr; i++){
   posu0 = i*axr_factor;
   posu1 = posu0+axu_factor;

   axr_config->index = (axr->index = i);
   nN_ = nN_np1[i]; uN = uN_np1[i];   
   kTi__ = kTi_[i];
   GammaN = GammaN_n[i];
   dtomega = dt_M*kTi__/nisp->D[0][i];
   dtomegai = dtomega*nN_;
   one_coll_G = 1.0/(1.0 + dtomegai);
   one_coll_Gammai = (i == Nr)?
     (1.0/(1.0 + dtomegai + dt*Gammai_W)):
     one_coll_G;
   G_eff = Ve[i]*(ni_np1[i]-ne[i]);
   ri = axr_config->r[i];
   d1 = dt*Vx*2.0*ri/(R*R);
   i_wall_rate = e_wall_rate = wall_rate = 0.0;
   if (i == Nr){
     e_wall_rate = electron_wall_rate*drN;
     i_Wall_Rate = i_wall_rate = ion_wall_rate*drN;
     wall_rate = (i_wall_rate - e_wall_rate);
     Gammae[i] -= e_wall_rate;
   }
   flux = (G_eff - ((Gammai_n[i] - i_wall_rate)  - Gammae[i]));
   Er_np1[i] = (
		Er_n[i] + dtq_eps0*flux
		)/(1.0 - d1);
   psum_ = dtomega*GammaN*ni_np1[i];
   Gammai_np1[i] = one_coll_Gammai*(Gammai_n[i] + psum_ + 
				    dtqi_M*ni_np1[i]*Er_np1[i] -
				    dt_M*(
					  ni_np1[i+1]*kTi_[i+1] - ni_np1[i-1]*kTi_[i-1]
					  )*one_del_2r[i]
				    );
   psum = dtomegai*Gammai_np1[i] - psum_;
   for(j=0; j<2*Nexp; j+=2){
     sp = exsp[j];
     Gsp = exsp[j+1];
     dtom = dt/sp->mass;
     dtomega =  dtom*kTi__/sp->D[0][i];
     dtomegai = nN_*dtomega;
     if (i == Nr){
       one_coll_ex = 1.0/(1.0 + dt*ex_W[j+1] + dtomegai);
     } else {
       one_coll_ex = 1.0/(1.0 + dtomegai);
     }
     psum_ = dtomega*sp->rho_np1[i]*GammaN;
     Gsp->rho_np1[i] = one_coll_ex*(
				    Gsp->rho_n[i] + psum_ + 
				    dtom*(sp->rho_np1[i]*sp->charge*Er_np1[i] 
					  - 
					  (
					   kTi_[i+1]*sp->rho_np1[i+1] -
					   kTi_[i-1]*sp->rho_np1[i-1]
					   )*one_del_2r[i]
					  )
				    );
     psum += (dtomegai*Gsp->rho_np1[i] - psum_);
   }
   /*
     Momentum source terms for neutrals
   */
   dtom = dt/nNsp->mass;
   GammaN_np1[i] = GammaN_n[i] - dtom*TWO_THIRDS*q0*(uN_np1[i+1]-uN_np1[i-1])*one_del_2r[i] + psum;
   /*

     Now put terms into Xterm, Eterm, rterm
     Assume Xterm has all collision terms in it at this point
     (except for the one with ddu_log_eta).
     update Xr, Xz

   */
  Er_ = Er_np1[i]; 
  Ez_ = *Ez_np1;
  nN_ = nN_np1[i]; uN = ONE_THIRD*uN_n[i]; /* uN = kT/2e */
  dd_2r = one_del_2r[i];
  for(j=1, pos = posu1; j<=(Nu); j++, pos += axu_factor){
    Xr_ = Xr_np1[pos]; Xz_ = Xz_np1[pos]; Psp = Ps_[pos];
    ddu_log_n = (
		 log_n[pos+axu_factor] - log_n[pos-axu_factor]
		   )*one_del_2u[j] ;
    ddu_log_eta = ddu_log_n + .5*one_u[j];
    ddr_log_eta = (
		   log_n[pos+axr_factor] - log_n[pos-axr_factor]
		   )*dd_2r;
    EdotX = (Er_*Xr_+Ez_*Xz_);
    Xterm_ = Xterm[pos] - (
		 del_elastic[j]*(nN_*u[j]-uN)*ddu_log_eta
	         )  - (
		      Xr_*Vs_[pos]*ddr_log_eta
		      ) + (
			   Vs_[pos]*EdotX*ddu_log_eta
			   );
     rterm = - (
		     v[j]*(
			   Psp*ddr_log_eta 
			   + (
			      Ps_[pos+axr_factor] - Ps_[pos-axr_factor]
			      )*dd_2r
			   )
		     );
     Eterm = v[j]*(
		   (Ps_[pos+axu_factor] - Ps_[pos-axu_factor])*one_del_2u[j] +
		   Psp*ddu_log_n
		   );
     Xr_np1[pos] = (Xr_ + dt*(Er_*Eterm + rterm))/(1.0 - dt*Xterm_);
     Xz_np1[pos] = (Xz_ + dt*(Ez_*Eterm))/(1.0 - dt*Xterm_);
     Xs = SQR(Xr_np1[pos])+SQR(Xz_np1[pos]);
     /*

       Test for violation of maximum anisotropy

       This probably means that the timestep is too large.
       Issue a time step violation.
       It's probably just a start-up problem, anyway.
       Should probably test for excess of something other than 1.0

       Alternatively:

       Limit Xr_ and Xz_ so that the rest of the computation will proceed
       without floating point exceptions. Limiting is first-order accurate
       in this context because each of the above rate terms goes to zero as 
       X goes to 1. 

     */
     if (Xs > POSCOL_Xs_LIMIT) {
       /*
       fprintf(stderr,"[%i %i]: Xs: %8.5e dt: %8.5e \n", i, j, Xs, dt);
       return(FALSE);
       */
       x = sqrt(Xs);
       Xr_np1[pos] /= x;
       Xz_np1[pos] /= x;
     }
     /*
       Now fix up boundaries
     */
     if (i == 1) {
       Xr_np1[pos-axr_factor] = -Xr_np1[pos];
       Xz_np1[pos-axr_factor] = Xz_np1[pos];
     }
  }
  Xr_np1[posu0] = Xrsp->bc[axu->sequence].Sl*Xr_np1[posu1];
  Xz_np1[posu0] = Xzsp->bc[axu->sequence].Sl*Xz_np1[posu1];
  eta_np1[posu0] = etasp->bc[axu->sequence].Sl*eta_np1[posu1];
  Xr_np1[pos] = Xr_np1[pos-axu_factor];
  Xz_np1[pos] = Xz_np1[pos-axu_factor];
 }

 axr->index = Nr+1;
/*
  boundary conditions
*/
 for(j=0; j<2*Nexp; j+=2){
   sp = exsp[j];
   Gsp = exsp[j+1];
   Gsp->rho_np1[Nr+1] = Gsp->bc[axr_config->sequence].Sr*Gsp->rho_np1[Nr];
   Gsp->rho_np1[0] = Gsp->bc[axr_config->sequence].Sl*Gsp->rho_np1[1];

 }
 Er_BC_V = (ni_np1[Nr]-ne[Nr])*(q0/eps0)*((R+axr_config->r[Nr])*(R-axr_config->r[Nr])/R);
 Er_np1[Nr+1] = Er_BC_V + Er_BC_S*Er_np1[Nr];
 Gammai_np1[0] = Gammaisp->bc[axr_config->sequence].Sl*Gammai_np1[1];
 Gammai_np1[Nr+1] = Gammaisp->bc[axr_config->sequence].Sr*Gammai_np1[Nr];
 Gammae[0] = -Gammae[1]; Gammae[Nr+1] = Gammae[Nr];
 nN_np1[0] = nN_np1[1];
 nN_np1[Nr+1] = nN_np1[Nr];
/* fix up corners */
 Xr_np1[0] = Xr_np1[r_R] = 0.0;
 Xr_np1[u_infty] = -Xr_np1[u_infty+axr_factor];
 Xr_np1[rusize-1] = Xr_np1[rusize-1-axr_factor];
 Xz_np1[0] = Xz_np1[r_R] = 0.0;
 Xz_np1[u_infty] = Xz_np1[u_infty+axr_factor];
 Xz_np1[rusize-1] = Xz_np1[rusize-1-axr_factor];

 return(TRUE);
}
/************************************************************

  sources()

************************************************************/
bool sources(set, dt)
MD_FCT_species_set_ptr set;
double dt;
{
  double vu, maxE, Xs, tmpdt;
  double absEr, maxEr, vr, Ez, Ti_, vj;
  double a, x, eta_, X_r, X_z, rhox, nN_, uN, ion_elastic_;
  double r1,r2;
  static double r, du;
  dblptr eta, Xr, Xz, ni, Gammai, Er;
  unsigned int i, j, pos, posu0, posu1;
  static unsigned int pnt, Dcnt = 0;
  static bool First_Time = true, first_time = true;
  static string gstring = NULL;
  static dblptr eta_r_v, eta_u_v, Xr_u_ad_v, Xr_r_ad_v, Xr_u_con_v, Xr_r_con_v;
  bool result;

if (First_Time){
  MD_LOGFILE = LOGFILE;
  MD_Courant_retry_factor = .7;
  MD_Courant_update_factor = .9;
  MD_Courant = .1;
  gstring = get_memory(gstring, 50, char);

  Nexp = 0;
  for(i=0; i<Nex; i++){
    if (Ex_products[i] != NULL) {Nexp++;}
  }
  exsp = get_memory(exsp, 2*Nexp, MD_FCT_species_ptr);
  ex_W = get_memory(ex_W, 2*Nexp, double);
  j=0;
  for(i=0; i<Nex; i++){
    if (Ex_products[i] != NULL){
      gstring = strcpy(gstring, "n");
      gstring = strcat(gstring,Ex_products[i]);
      exsp[j] = MD_get_spptr(gstring, set);
      gstring = strcpy(gstring,"Gamma");
      gstring = strcat(gstring,Ex_products[i]);
      exsp[j+1] = MD_get_spptr(gstring, set);
      exsp[j+1]->vel = exsp[j]->vel;
      j += 2;
    }
  }
  r_R = (Nr+1)*axr_factor;
  rusize = (Nr+2)*(Nu+2);
  u_infty = (Nu+1)*axu_factor;

  eta_TBC = get_memory(eta_TBC, (Nr+2), double);
  Ps = get_memory(Ps, rusize, double);
  Vs = get_memory(Vs, rusize, double);
  Xterm = get_memory(Xterm, rusize, double);
  log_n = get_memory(log_n, rusize, double);
  t1 = get_memory(t1, (Nu+2), double);
  t2 = get_memory(t2, (Nu+2), double);
  Ve = get_memory(Ve, (Nr+2), double);
  C = get_memory(C, (Nr+2), double);
  exch = get_memory(exch, (Nr+2), double);
  comp = get_memory(comp, (Nr+2), double);
  cond = get_memory(cond, (Nr+2), double);
  kTi_ = get_memory(kTi_, (Nr+2), double);

  ion_elastic = qi/(m_N*mui);

  free(Gammaisp->vel->velocity);
  free(GammaNsp->vel->velocity);
  free(uNsp->vel->velocity);

  eta_u_v = eta_u_vel->velocity;
  Xr_u_ad_v = Xr_u_ad_vel->velocity;

  free(Xr_r_ad_vel->velocity);
  free(Xz_u_ad_vel->velocity);
  free(Xz_r_ad_vel->velocity);
  free(Xz_u_con_vel->velocity);
  free(Xz_r_con_vel->velocity);

  
  Xz_r_ad_vel->velocity = Xr_r_ad_v = Xr_r_ad_vel->velocity = eta_r_v = eta_r_vel->velocity;
  Xz_u_ad_vel->velocity = Xr_u_ad_vel->velocity;
  Xz_r_ad_vel->velocity = Xr_r_ad_vel->velocity;

  Xz_u_con_vel->velocity = Xr_u_con_v = Xr_u_con_vel->velocity;
  Xz_r_con_vel->velocity = Xr_r_con_v = Xr_r_con_vel->velocity;

  Gammaisp->vel->velocity = nisp->vel->velocity;
  uNsp->vel->velocity = GammaNsp->vel->velocity = nNsp->vel->velocity;
/*
  get any u-dependent arrays
*/
  delel = 2*m_elec/m_N;
  qi_M = qi/m_N;
  kTi = k*Ti;
  kTi_e = kTi/q0;
  kTi_M = kTi/m_N;
  get_elastic();
  get_inelastics(set);
  get_step_control();
  /* get Ve */
  Vx = .0102*STEOM*sqrt(uMAX);
  for(i=1; i<=(Nr+1); i++){
    Ve[i] = Vx*SQR(axr_config->r[i]/R);
  }
  Ve[0] = -Ve[1];
  free(Ersp->vel->velocity);
  Ersp->vel->velocity = Ve;
  Er_BC_S = 1.0;

  pnt = Nu/10;
  Xscnt = 0;
  du = axu->r[axu_factor*(Nu+1)] - axu->r[axu_factor*Nu];
  r = du/(axu->r[axu_factor*Nu] - axu->r[(Nu-pnt)*axu_factor]);
  one_u_amb = q0/(Ptorr*TORR_TO_BAR*1.5);

  First_Time = false;
}
/*
  Calculate velocities based on sp->rho
*/
 eta = etasp->rho;
 Xr = Xrsp->rho;
 Xz = Xzsp->rho;
 ni = nisp->rho;
 Gammai = Gammaisp->rho;
 Er = Ersp->rho;
 Ez = Ezsp->rho[0];
 for(axr->index=i=1; i<= (Nr); axr->index = ++i){
   posu0 = i*axr_factor; posu1 = posu0 + axu_factor;
   uN = ONE_THIRD*uNsp->rho[i];
   nN_ = nNsp->rho[i];
   for(pos = posu1, j=1; j<= (Nu+1);  ++j, pos += axu_factor){
     vj = v[j];
     X_r = Xr[pos]; X_z = Xz[pos]; eta_ = eta[pos];
     Xs = (X_r*X_r+X_z*X_z);
     vr = vj*X_r;
     vu = -vj*(X_r*Er[i]+X_z*Ez);
     if (Xs > 2.0) {
       fprintf(stderr,"Xs[%u,%u]: %8.5e %8.5e %8.5e %8.5e\n",i,j,Xs, eta_, X_r, X_z);
       Xscnt++;
     }
     if (Xscnt > 10){ fprintf(stderr,"Failed\n"); exit(1);}
     eta_r_v[pos] = vr;
     eta_u_v[pos] = vu - del_elastic[j]*(nN_*u[j]-uN);
     Xr_u_ad_v[pos] = vu;
     x = (get_Vterm(Xs)-1.0);
     Xr_r_con_v[pos] = x*vr;
     Xr_u_con_v[pos] = x*vu;
     
     if (i == 1) {
       eta_u_v[pos - axr_factor] = eta_u_v[pos];  
       eta_r_v[pos - axr_factor] = -eta_r_v[pos];
       Xr_u_ad_v[pos - axr_factor] = Xr_u_ad_v[pos];
       Xr_u_con_v[pos - axr_factor] = Xr_u_con_v[pos];  
       Xr_r_con_v[pos - axr_factor] = -Xr_r_con_v[pos];
     }
     if (i == Nr){
      eta_u_v[pos + axr_factor] = eta_u_v[pos];
      eta_r_v[pos + axr_factor] = eta_r_v[pos];
      Xr_u_ad_v[pos + axr_factor] = Xr_u_ad_v[pos];
      Xr_u_con_v[pos + axr_factor] = Xr_u_con_v[pos];  
      Xr_r_con_v[pos + axr_factor] = Xr_r_con_v[pos];
    } 
  }
  Xr_u_con_v[posu0] = -Xr_u_con_v[posu1];
  eta_u_v[posu0] = -eta_u_v[posu1];
  Xr_u_ad_v[posu0] = -Xr_u_ad_v[posu1];
  Xr_r_con_v[posu0] = 0.0;
  eta_r_v[posu0] = 0.0;
  pos = i*axr_factor + Nu*axu_factor;
  x = eta[pos]*v[Nu]/(eta[pos-pnt*axu_factor]*v[Nu-pnt]);
  if (x > 0) {
    eta_TBC[i] = pow(x,r);
  } else {
    fprintf(stderr,"Error at [%i, %i] x = %8.5e\n",i,Nu, x);
  }
  nNsp->vel->velocity[i] = GammaNsp->rho[i]/nNsp->rho[i];
 }

 nNsp->vel->velocity[0] = -nNsp->vel->velocity[1];
 nNsp->vel->velocity[Nr+1] = -nNsp->vel->velocity[Nr];

 for(i=1; i<= (Nr+1); i++){
   for(j=0; j<2*Nexp; j+=2){
     rhox = exsp[j]->rho[i];
     exsp[j]->vel->velocity[i] = (rhox > 0.0)?(exsp[j+1]->rho[i]/rhox):0.0;
     if (i == 1) {
       exsp[j]->vel->velocity[0] = -exsp[j]->vel->velocity[1];
     }
   }
   nisp->vel->velocity[i] = (ni[i] != 0.0)?(Gammai[i]/ni[i]):0.0;
 }
 nisp->vel->velocity[0] = -nisp->vel->velocity[1]; 
/*
  get excited species collision terms
*/
 for(j=0; j<2*Nexp; j+=2){
   get_heavy_wall_terms(exsp[j], &(ex_W[j]), exsp[j+1], &(ex_W[j+1]));
 }
/*
  get ion wall collision terms
*/
 get_heavy_wall_terms(nisp, &ni_W, Gammaisp, &Gammai_W );
/*
  Advance variables by updating source terms
*/
 result = MD_FCT_Richardson(set, advance_sources, dt);
 if (!result) { return(FALSE); }

/*
  Determine new time step

  get maximum Er
  Take conservative (noncollisional) time step
  restriction from (du/E*v) and (dr/v).
  Then do ion contribution.

*/
maxEr = 0.0; 
for(i=1; i<=Nr; i++){
  absEr = ABS(Er[i]);
  maxEr = max(maxEr, absEr);
}
maxE = maxEr + ABS(Ez);

MD_Source_dt = min(dtkinetic, mindu_v/maxE);

a = ion_a1*maxEr + ion_a2;

ion_elastic_ = ion_elastic*nN;
tmpdt = (ion_elastic_ + sqrt(SQR(ion_elastic_) + 4*a))/(2*a);

MD_Source_dt = min(MD_Source_dt, tmpdt);
/*
  print status and log data
*/
 if (MD_FCT_step == full){
   Dcnt++;
   fprintf(stderr,"MD_Source_dt: %8.5e dt: %8.5e total time: %8.5e ",MD_Source_dt, dt, MD_Total_time);
   fprintf(stderr,"Ez: %8.5e I: %8.5e\n", Ez,I);
   /* occasionally update the diffusion coefficients */
   if (Dcnt > 32) {
     for(i=1; i<=Nr; i++){
       Ti_ = get_Ti(uNsp->rho[i],nNsp->rho[i]);
       nisp->D[0][i] = Di(Ti_);
       for(j=0; j<2*Nexp; j+=2){
	 exsp[j]->D[0][i] = Dex(Ti_);
       }
     }
     Dcnt = 0;
   }
   /* do filtering to get averages */
   r1 = dt/(TAU_AV+dt); r2 = TAU_AV/(dt+TAU_AV);
   for(i=0;  i<=Nr+1; i++){
     if (first_time){
       Gammae_av[i] = Gammae[i];
     } else {
       Gammae_av[i] = r1*Gammae[i] + r2*Gammae_av[i];
       Er_av[i] = r1*Er[i] + r2*Er_av[i];
     }
   }
   first_time = false;
 }
 return(TRUE);
}

