/************************************************************

  sources.c

  source terms for poscol simulation

************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "basics.h"
#undef pi
#include "MD_FCT_defs.h"
#include "MD_FCT.h"
#include "MD_FCT_utils.h"
#include "poscol.h"
#include "special.h"
#include "collisions.h"
#define pi 3.14159265
#define SRC_LIMITER
#define VEL_LIMITER
#define PSU

#define GET_X_GAMMA(xs) ((xs>=1.0)?(sqrt(xs)):((35.0 + xs*(92.0+xs*17.0))/(105.0 + xs*(24.0+xs*15.0))))
#define GET_VTERM(xs) ((xs>=1.0)?.5*(3.0-sqrt(xs)):((126.0 + 18.0*xs)/(105.0 + xs*(24.0+xs*15.0))))
#define GET_PTERM(xs) ((xs>=1.0)?(0.0):((70.0 -xs*(68.0+xs*2.0))/(105.0 + xs*(24.0+xs*15.0))))

MD_FCT_species_ptr etasp, Grsp, Gzsp, nisp, Gammaisp, Ersp, Ezsp, Nsp;
static MD_FCT_species_ptr *exsp = NULL;
unsigned int axu_factor, axr_factor, rusize, u_infty, r_R, vr_offset, vu_offset;
geometry_ptr phase_space, configuration_space;
axis_ptr axr_config;
dblptr elastic = NULL, del_elastic = NULL, u_h = NULL, one_u_h = NULL;
dblptr ex_W = NULL, ve = NULL, vs = NULL;
dblptr eta_n, eta_np1, Gr_n, Gr_np1, Gz_n, Gz_np1, ne = NULL;
double delel, kTi_e, kTi_M, Vx, qi_M, ion_a1, ion_a2;
double dtkinetic, dr_W;
dblptr nuM = NULL, XS = NULL;
dblptr u_av = NULL, Gammae = NULL;
dblptr eta_TBC = NULL, Gr_TBC = NULL, Gz_TBC = NULL;

static dblptr C = NULL;
static dblptr t1 = NULL, Ps = NULL, QE = NULL, Ve = NULL;
static double ion_elastic;
static double Ez, Er_BC, kTi;
static unsigned int u_ctr = 0, u_freq = 16, Nexp = 0;
static double ni_W, Gammai_W;
double teoms, mindu_v, ion_sound_speed;

void get_heavy_wall_terms(MD_FCT_species_ptr, dblptr, MD_FCT_species_ptr, dblptr);
void print_row();
double get_X_gamma(double);
double get_Pterm(double);
double get_vterm(double);
/***************************************************

  print_neighborhood(y,pos)
  
  prints MxM neighborhood of 2-d array y

***************************************************/
void print_neighborhood(y,M)
dblptr y;
unsigned int M;
{
int i,j,rpos,upos,N;
unsigned int pos;
char outstr[100];

N = M/2;
pos = MD_get_position(phase_space);
for(j=N; j>=-N; j--){
  upos = axu->index + j;
  for(i=-N; i<=N; i++){
    rpos = (axr->index + i);
    if ((rpos > Nr+1)||(rpos < 0)||(upos > Nu+1)||(upos < 0)){
      sprintf(outstr, " ************ ");
    } else {
      sprintf(outstr," % 9.5e ",y[pos+i*axr_factor+j*axu_factor]);
    }
    fprintf(LOGFILE,"%s",outstr);
  }
  fprintf(LOGFILE,"\n");
}
return;
}
/************************************************************

  eta_OutFlow_u()

  Outflow boundary condition using u_char, v_sonic_u.

************************************************************/
void eta_OutFlow_u(bc)
MD_FCT_bc *bc;
{

  bc->vchar_r = 0.0;
  bc->Lchar_r = 0.0;
  bc->Sr = eta_TBC[axr->index]; bc->Vr = 0.0;
  bc->rho_amb_r = 0.0;
}
/************************************************************

  Gr_OutFlow_u()

  Outflow boundary condition using u_char, v_sonic_u.

************************************************************/
void Gr_OutFlow_u(bc)
MD_FCT_bc *bc;
{

  bc->Sr = Gr_TBC[axr->index]; 
  bc->Vr = 0.0;
  bc->vchar_r = 0.0;
  bc->Lchar_r = 0.0;
  bc->rho_amb_r = 0.0;
}
/************************************************************

  Gz_OutFlow_u()

  Outflow boundary condition using u_char, v_sonic_u.

************************************************************/
void Gz_OutFlow_u(bc)
MD_FCT_bc *bc;
{

  bc->Sr = Gz_TBC[axr->index]; 
  bc->Vr = 0.0;
  bc->vchar_r = 0.0;
  bc->Lchar_r = 0.0;
  bc->rho_amb_r = 0.0;
}
/************************************************************

  eta_wall()

  wall boundary condition for eta

************************************************************/
void eta_wall(bc)
MD_FCT_bc *bc;
{

  bc->Sr = SBC[axu->index];
  bc->Vr = 0.0;

  bc->Lchar_r = 0.0;
}
/************************************************************

  Gr_wall()

  wall boundary condition for Gr

************************************************************/
void Gr_wall(bc)
MD_FCT_bc *bc;
{
  bc->Sr = SBC[axu->index];
  bc->Vr = 0.0;

  bc->Lchar_r = 0.0;
}
/************************************************************

  Gz_wall()

  wall boundary condition for Gz

************************************************************/
void Gz_wall(bc)
MD_FCT_bc *bc;
{
  bc->Sr = SBC[axu->index];
  bc->Vr = 0.0;
  bc->Lchar_r = 0.0;
}
/************************************************************

  Er_wall()

  wall boundary condition for Er

************************************************************/
void Er_wall(bc)
MD_FCT_bc *bc;
{
  bc->Sr = Er_BC;
  bc->Vr = 0.0;
  bc->Lchar_r = 0.0;
}
/************************************************************

     get_X_gamma(Xs)

     gets X/gamma according to approximate inversion of
     the definition by the elliptic expansion.

     Xs must be X*X;

************************************************************/
double get_X_gamma(Xs)
     double Xs;
{
  double result;

  if (Xs >= 1.0) {return(sqrt(Xs));}
  result = (35.0+Xs*(92.0+Xs*17.0))/(105.0 + Xs*(24.0+Xs*15.0));
  return(result);
}
/************************************************************

     get_Pterm(Xs)

     gets (1 - X/gamma) according to approximate inversion of
     the definition by the elliptic expansion.

     Xs must be X*X;

************************************************************/
double get_Pterm(Xs)
     double Xs;
{
  double result;

  if (Xs >= 1.0) {return(1.0-sqrt(Xs));}
  result = (70.0-Xs*(68.0+Xs*2.0))/(105.0 + Xs*(24.0+Xs*15.0));
  return(result);
}
/************************************************************

     get_vterm(Xs)

     gets (3(X/gamma)-1)/(2X*X) according to approximate inversion of
     the definition by the elliptic expansion.

     Xs must be X*X;

************************************************************/
double get_vterm(Xs)
     double Xs;
{
  double result;

  if (Xs >= 1.0) {return(.5*(3.0-sqrt(Xs)));}
  result = (126.0 + 18.0*Xs)/(105.0 + Xs*(24.0+Xs*15.0));
  return(result);
}
/************************************************************

  get_u_av()

  gets average energy based on eta->rho0[]

************************************************************/
void get_u_av()
{
  unsigned int i, j;
  unsigned int pos, posu1;

  for(i=1; i<=Nr; i++){
    axr->index = i;
    posu1 = i*axr_factor+axu_factor;
    for(j=1, pos = posu1; j<=Nu; j++, pos += axu_factor){
      t1[j] = etasp->rho0[pos]*axu->r[pos];
    }
    u_av[i] = MD_integral_dV(t1, axu, phase_space)/
      MD_integral_dVol(etasp->rho0, axu, phase_space);
  }
  u_av[0] = u_av[1];  u_av[Nr+1] = 0.0;
  return;
}
/************************************************************

  get_heavy_wall_terms()

  gets ni_W, Gammai_W, ni_BC, Gammai_BC for
  any heavy species

************************************************************/
void get_heavy_wall_terms(sp, n_W, Gammasp, Gamma_W)
MD_FCT_species_ptr sp, Gammasp;
dblptr n_W, Gamma_W;
{
  double d, x, c1, c2, E4, E4_, cs, x1, x2;
  double sound_speed, vr, n_BC;

  vr = sp->vel[Nr];
  sound_speed = sqrt((2.0*k*Ti)/sp->mass);
  d = (R - axr_config->r[Nr]);
  c1 = (vr+SQRT__PI_2*sound_speed);
  x1 = (d*ion_elastic)/c1;
  E4 = (c1>0.0)?En(x1,4):0.0;
  c2 = (vr-SQRT__PI_2*sound_speed);
  x2 = (d*ion_elastic)/c2;
  E4_ = (c2>0.0)?En(x2,4):0.0;
  n_BC = 6.0*(E4 + E4_) - 1.0;
  sp->bc[axr_config->sequence].Sr = n_BC;
  Gammasp->bc[axr_config->sequence].Sr = n_BC;
  cs = sound_speed*sound_speed;
  x = (E4 - E4_)/(sqrt_2*SQRT_PI*d*sound_speed);
  *n_W = 1.5*(vr*vr+PI_2*cs)*x;
  *Gamma_W = (vr*vr + 1.5*PI*cs)*x;
  return;
}
/************************************************************

  advance_sources()

************************************************************/
void advance_sources(set, dt)
double dt;
MD_FCT_species_set_ptr set;
{
MD_FCT_species_ptr sp, Gsp;
dblptr ni_n, ni_np1, Gammai_n,  Gammai_np1;
dblptr Ez_n, Ez_np1, Er_n, Er_np1;
unsigned int i, j, sl, sr;
unsigned int pos, posu0, posu1;
double dtq_eps0, dtkTi_M, dtqi_M;
double one_coll_G, one_coll_Gammai, one_coll_ni, one_coll_r, one_coll_z;
double G_eff, one_coll_ex, dtom, ddr, ddu, u, d1, d2;
double a, a_last, t2, Gr_, Gz_, eta_, Xs, X;
static bool First_Time = true;
static char sw;

if (First_Time) {
  C = get_memory(C, (Nr+2), double);
  sw = MD_Diffusion_re_use_Dkappa|MD_Diffusion_first_fractional_step;
  qi_M = qi/m_N;
  First_Time = false;
}
eta_n = etasp->rho_n;
eta_np1 = etasp->rho_np1;
Gr_n = Grsp->rho_n;
Gr_np1 = Grsp->rho_np1;
Gz_n = Gzsp->rho_n;
Gz_np1 = Gzsp->rho_np1;
ni_n = nisp->rho_n;
ni_np1 = nisp->rho_np1;
Gammai_n = Gammaisp->rho_n;
Gammai_np1 = Gammaisp->rho_np1;
Er_n = Ersp->rho_n;
Er_np1 = Ersp->rho_np1;
Ez_n = Ezsp->rho_n;
Ez_np1 = Ezsp->rho_np1;
dtq_eps0 = dt*q0/eps0;
dtkTi_M = dt*kTi_M; 
dtqi_M = dt*qi_M;

/************************************************************
  
  Get diffusion solution for finite Ti...take eta_n -> eta_np1

************************************************************/
MD_Diffusion(etasp, dt, axu, Du, sw);
/************************************************************

  put together the inelastic transport matrix
  (1 + de_excitation_dat*dt)/(1 + inelastic*dt)
  and solve it. Take eta_np1 -> eta_np1.

************************************************************/
Apply_inelastics(eta_np1, dt, true, true);
/*
  boundary terms for ions and excited species
*/
one_coll_ni = 1.0/(1.0+dt*ni_W);
ni_np1[Nr] *= one_coll_ni;
ni_np1[Nr+1] = nisp->bc[axr_config->sequence].Sr*ni_np1[Nr];
for(j=0; j<2*Nexp; j+=2){
  sp = exsp[j];
  Gsp = exsp[j+1];
  sp->rho_np1[Nr] *= 1.0/(1.0 + dt*ex_W[j]);
  sp->rho_np1[Nr+1] = sp->bc[axr_config->sequence].Sr*sp->rho_np1[Nr];
}
/************************************************************

  get parameters for ion wall collision rate
  get new ion concentrations from source terms

  include wall collision terms for Gr, Gz

  get C, D
  get Gr_np1
************************************************************/
for(i=1; i<=Nr; i++){
  posu0 = i*axr_factor;
  posu1 = posu0+axu_factor;
  axr_config->index = (axr->index = i);
  ne[i] = FOURPIEOM*MD_integral_dVol(eta_np1, axu, phase_space);
  for(j=1, pos = posu1; j<=(Nu+1); j++, pos += axu_factor){
    Gr_ = Gr_n[pos]; Gz_ = Gz_n[pos];
    eta_ = eta_np1[pos];
    if (eta_ > 0.0){
      Xs = (Gr_*Gr_+Gz_*Gz_)/(vs[j]*eta_*eta_);
    } else {
      Xs = 0.0; 
      eta_np1[pos] = 0.0;
    }
    XS[pos] = Xs;
#ifdef NO_PSU
    Ps[pos] = eta_np1[pos]*eom*GET_PTERM(Xs);
    QE[pos] = eta_np1[pos]*eom*Xs*GET_VTERM(Xs);
#endif
#ifdef PSU
    Ps[pos] = axu->r[pos]*eta_np1[pos]*eom*GET_PTERM(Xs);
    QE[pos] = eta_np1[pos]*eom;
#endif
    if (i == Nr){
      sr = pos+axr_factor;
      sl = sr - r_R;
      Ps[sr] = SBC[j]*Ps[pos];
      Ps[sl] = etasp->bc[axr->sequence].Sl*Ps[sl+axr_factor];
      XS[sr] = XS[pos];
      XS[sl] = XS[sl+axr_factor];
    }      
  }
  Ps[posu0] = etasp->bc[axu->sequence].Sl*Ps[posu1];
  XS[posu0] = 0.0;
}
axr->index = Nr+1;
ne[Nr+1] = FOURPIEOM*MD_integral_dVol(eta_np1, axu, phase_space);
ne[0] = ne[1];
/*
  Get Ez, Er, Gr, Gammai
*/
for(axr->index=1; axr->index<=Nr+1; axr->index++){
  C[axr->index] = MD_integral_dVol(Gz_n, axu, phase_space);
  Gammae[axr->index] = FOURPIEOM*MD_integral_dVol(Gr_n,axu, phase_space);  
}
C[0] = C[1];
I = -FOURPIEOM*2.0*pi*q0*MD_integral_dVol(C, axr_config, configuration_space);
(*Ez_np1) = ((*Ez_n) + dt*(I0 - I)/Cl);
one_coll_G = 1.0/(1.0 + dt*ion_elastic);
for(i=1; i<=Nr; i++){
  posu0 = i*axr_factor;
  posu1 = posu0+axu_factor;
  axr_config->index = (axr->index = i);

  one_coll_Gammai = (i == Nr)?(1.0/(1.0 + dt*(ion_elastic + Gammai_W))):one_coll_G;
  G_eff = Ve[i]*(ni_np1[i]-ne[i]);
  d1 = dt*Vx*2.0*axr_config->r[i]/(R*R);
  d2 = dtq_eps0*dt*eom*ne[i];
  Er_np1[i] = (
		    Er_n[i] + dtq_eps0*(G_eff - (Gammai_n[i] - Gammae[i]))
		    )/
		 (1.0 - d1 + d2);
  /*
    Transport the ions
  */
  Gammai_np1[i] = one_coll_Gammai*(Gammai_n[i] +
				   dtqi_M*ni_np1[i]*Er_np1[i] -
				   dtkTi_M*MD_d_dr(ni_np1, axr_config, i)
				   );
  /* 
     Transport all of the excited species 
  */
  for(j=0; j<2*Nexp; j+=2){
    sp = exsp[j];
    Gsp = exsp[j+1];
    dtom = dt/sp->mass;
    if (i == Nr){
      one_coll_ex = 1.0/(1.0 + dt*ex_W[j+1] + dtom*(kTi/sp->D[0][i])*nN);
    } else {
      one_coll_ex = 1.0/(1.0 + dtom*(kTi/sp->D[0][i])*nN);
    }
    Gsp->rho_np1[i] = one_coll_ex*(Gsp->rho_n[i] + 
				   dtom*(sp->rho_np1[i]*sp->charge*Er_np1[i] - 
					 kTi*MD_d_dr(sp->rho_np1, axr_config, i))
				   );
  }
#ifdef NO_PSU
  a_last = Ps[posu1]*one_u_h[1];
#endif
#ifdef PSU
  a_last = sqrt(Ps[posu0]*Ps[posu1]);
  /*  a_last = .5*(Ps[posu0] + Ps[posu1]);*/
#endif
  for(j=1, pos = posu1; j<=(Nu); j++, pos += axu_factor){
    axu->index = j;
    one_coll_r = 1.0/(1.0 + dt*nuM[pos]);
    one_coll_z = one_coll_r;

#ifdef NO_PSU
    a = sqrt(Ps[pos]*Ps[pos+axu_factor]*one_u_h[j]*one_u_h[j+1]);
    /*    a = .5*(Ps[pos]*one_u_h[j] + Ps[pos+axu_factor]*one_u_h[j+1]);*/
    ddr = (u = axu->r[pos])*MD_d_dr(Ps, axr, pos) ;
    ddu = u*u_h[j]*2.0*(a - a_last)/(axu->r[pos+axu_factor]-axu->r[pos-axu_factor]);
#endif
#ifdef PSU
    a = sqrt(Ps[pos]*Ps[pos+axu_factor]);
    /*    a = .5*(Ps[pos]+Ps[pos+axu_factor]);*/
    ddr = MD_d_dr(Ps,axr,pos);
    ddu = 2.0*(a - a_last)/(axu->r[pos+axu_factor]-axu->r[pos-axu_factor]);
#endif
    t2 = ddu - QE[pos];
    a_last = a;

    Gr_np1[pos] = one_coll_r*(Gr_n[pos]  
			      + dt*(
				    Er_np1[i]*t2  -  ddr
				    )
			       );
    Gz_np1[pos] = one_coll_z*(Gz_n[pos] + dt*(*Ez_np1)*t2);
#ifdef SRC_LIMITER    
    eta_ = eta_np1[pos];
    if (eta_ > 0.0){
      Xs = (SQR(Gr_np1[pos]) + SQR(Gz_np1[pos]))/(vs[j]*SQR(eta_));
      if (Xs > 1.0) {
	X = sqrt(Xs);
	Gr_np1[pos] /= X;
	Gz_np1[pos] /= X;
	Xs = 1.0;
      }
    } else {
      Xs = 0.0;
      Gr_np1[pos] = Gz_np1[pos] = 0.0;
    }
    XS[pos] = Xs;
#endif
    if (i == 1) {
      Gr_np1[pos-axr_factor] = -Gr_np1[pos];
      Gz_np1[pos-axr_factor] = Gz_np1[pos];
      XS[pos-axr_factor] = XS[pos];
    }
    if (i == Nr) {
      Gr_np1[pos+axr_factor] = SBC[j]*Gr_np1[pos];
      Gz_np1[pos+axr_factor] = SBC[j]*Gz_np1[pos];
      XS[pos+axr_factor] = XS[pos];
    }
  }
  Gz_np1[pos] = Gz_TBC[i]*Gz_np1[pos-axu_factor];
  Gr_np1[pos] = Gr_TBC[i]*Gr_np1[pos-axu_factor];
  Gr_np1[posu0] = Gz_np1[posu0] = 0.0;
  Nsp->rho_np1[i] = nN;
  XS[pos] = XS[pos-axu_factor];
}
Nsp->rho_np1[0] = nN;
Nsp->rho_np1[Nr+1] = 0.0;
axr->index = Nr+1;
/*
  boundary conditions
*/
 for(j=0; j<2*Nexp; j+=2){
   sp = exsp[j];
   Gsp = exsp[j+1];
   Gsp->rho_np1[Nr+1] = Gsp->bc[axr_config->sequence].Sr*Gsp->rho_np1[Nr];
   Gsp->rho_np1[0] = Gsp->bc[axr_config->sequence].Sl*Gsp->rho_np1[1];

 }
Er_np1[Nr+1] = (Er_BC = (axr_config->r[Nr]/axr_config->r[Nr+1]))*Er_np1[Nr];
Gammai_np1[0] = Gammaisp->bc[axr_config->sequence].Sl*Gammai_np1[1];
Gammai_np1[Nr+1] = Gammaisp->bc[axr_config->sequence].Sr*Gammai_np1[Nr];
Gammae[0] = -Gammae[1];
/* fix up corners */
Gr_np1[0] = Gr_np1[r_R] = 0.0;
Gr_np1[u_infty] = -Gr_TBC[1]*Gr_np1[u_infty+axr_factor-axu_factor];
Gr_np1[rusize-1] = SBC[Nu+1]*Gr_TBC[Nr]*Gr_np1[rusize-1-axr_factor-axu_factor];
Gz_np1[0] = Gz_np1[r_R] = 0.0;
Gz_np1[u_infty] = Gz_np1[u_infty+axr_factor];
Gz_np1[rusize-1] = SBC[Nu+1]*Gz_np1[rusize-1-axr_factor];
/*
  get average energy every u_freq steps
*/
if (MD_FCT_step == half){
  if (!u_ctr){ get_u_av();}
  u_ctr = (u_ctr+1)%u_freq;
}
return;
}
/************************************************************

  sources()

************************************************************/
void sources(set, dt)
MD_FCT_species_set_ptr set;
double dt;
{
double vu, maxE, Xs, tmpdt, maxEs; 
double absEr, maxEr, vr, X;
double a, b, c, d, x, r, du, eta_, G_r, G_z, rhox;
dblptr eta, Gr, Gz, ni, Gammai, Er;
unsigned int i, j, pos, vu_pos, vr_pos, ip, jp, posu0, posu1;
static unsigned int pnt;
static bool First_Time = true;
static string gstring = NULL;
bool fail;

if (First_Time){
  MD_LOGFILE = LOGFILE;
  MD_Courant_retry_factor = .7;
  MD_Courant_update_factor = .9;
  MD_Courant = .25;
  gstring = get_memory(gstring, 50, char);
  etasp = MD_get_spptr("eta",set);  
  Grsp = MD_get_spptr("Gr",set);  
  Gzsp = MD_get_spptr("Gz",set);  
  nisp = MD_get_spptr("ni",set);
  Gammaisp = MD_get_spptr("Gammai",set);
  Ersp = MD_get_spptr("Er",set);
  Ezsp = MD_get_spptr("Ez",set);
  Nsp = MD_get_spptr("nN", set);

  Nexp = 0;
  for(i=0; i<Nex; i++){
    if (Ex_products[i] != NULL) {Nexp++;}
  }
  exsp = get_memory(exsp, 2*Nexp, MD_FCT_species_ptr);
  ex_W = get_memory(ex_W, 2*Nexp, double);
  j=0;
  for(i=0; i<Nex; i++){
    if (Ex_products[i] != NULL){
      gstring = strcpy(gstring, "n");
      gstring = strcat(gstring,Ex_products[i]);
      exsp[j] = MD_get_spptr(gstring, set);
      gstring = strcpy(gstring,"Gamma");
      gstring = strcat(gstring,Ex_products[i]);
      exsp[j+1] = MD_get_spptr(gstring, set);
      exsp[j+1]->vel = exsp[j]->vel;
      j += 2;
    }
  }
  phase_space = etasp->grid;
  axu_factor = axu->mul;
  axr_factor = axr->mul;
  r_R = (Nr+1)*axr_factor;
  rusize = (Nr+2)*(Nu+2);
  u_infty = (Nu+1)*axu_factor;
  vu_offset = axu->sequence*rusize;
  vr_offset = axr->sequence*rusize;
  ne = get_memory(ne, (Nr+2), double);
  eta_TBC = get_memory(eta_TBC, (Nr+2), double);
  Gr_TBC = get_memory(Gr_TBC, (Nr+2), double);
  Gz_TBC = get_memory(Gz_TBC, (Nr+2), double);
  Ps = get_memory(Ps, (Nu+2)*(Nr+2), double);
  QE = get_memory(QE, (Nu+2)*(Nr+2), double);
  XS = get_memory(XS, (Nu+2)*(Nr+2), double);
  t1 = get_memory(t1, (Nu+2), double);
  Ve = get_memory(Ve, (Nr+2), double);

  configuration_space = nisp->grid;
  axr_config = MD_get_axptr("r", configuration_space);

  teoms = teom*teom;
  ion_elastic = qi/(m_N*mui);

  free(Gammaisp->vel);
  free(Gzsp->vel); 
  Gzsp->vel = Grsp->vel;

  Gammaisp->vel = nisp->vel;
  u_av = get_memory(u_av, (Nr+2), double);
  Gammae = get_memory(Gammae, (Nr+2), double);
/*
  get any u-dependent arrays
*/
  delel = 2*m_elec/m_N;
  kTi = k*Ti;
  kTi_e = kTi/q0;
  kTi_M = kTi/m_N;
  get_elastic();
  get_inelastics(set);
  get_step_control();
  get_u_av();
  /* get Ve */
  Vx = .05*steom*sqrt(uMAX);
  for(i=0; i<=(Nr+1); i++){
    Ve[i] = Vx*SQR(axr_config->r[i]/R);
  }
  pnt = 10;
  First_Time = false;
}
/*
  Calculate velocities based on sp->rho
*/
eta = etasp->rho;
Gr = Grsp->rho;
Gz = Gzsp->rho;
ni = nisp->rho;
Gammai = Gammaisp->rho;
Er = Ersp->rho;
Ez = Ezsp->rho[0];
fail = false;
for(axr->index=i=1; i<= (Nr); axr->index = ++i){
  posu0 = i*axr_factor; posu1 = posu0 + axu_factor;
  for(pos = posu1, axu->index=j=1; j<= (Nu+1); axu->index = ++j, pos += axu_factor){
    pos = MD_get_position(phase_space);
    G_r = Gr[pos]; G_z = Gz[pos]; eta_ = eta[pos];
    if (eta_ > 0.0){
      Xs = (G_r*G_r+G_z*G_z)/(eta_*eta_*vs[j]);
      vr = G_r/eta_;
      vu = -(G_r*Er[i]+G_z*Ez)/eta_;
#ifdef VEL_LIMITER
      if (Xs > 1.0) {
	X = sqrt(Xs);
	vr /= X;
	vu /= X;
	Xs = 1.0;
      }
#endif
    } else {
      if (eta_ < 0.0){
	ip = i; jp = j;
	fail = true;
      }
      vr = vu = 0.0; Xs = 0.0;
    }
    XS[pos] = Xs;
    vu_pos = vu_offset + pos;
    vr_pos = vr_offset + pos;
    etasp->vel[vr_pos] = vr;
    etasp->vel[vu_pos] = vu - del_elastic[j];
    x = GET_VTERM(Xs);
    Grsp->vel[vr_pos] = x*vr;
    Grsp->vel[vu_pos] = x*vu;

    if (i == 1) {
      etasp->vel[vu_pos - axr_factor] = etasp->vel[vu_pos];  
      etasp->vel[vr_pos - axr_factor] = -etasp->vel[vr_pos];
      Grsp->vel[vu_pos - axr_factor] = Grsp->vel[vu_pos];  
      Grsp->vel[vr_pos - axr_factor] = -Grsp->vel[vr_pos];
      XS[pos-axr_factor] = XS[pos];
    }
    if (i == Nr){
      etasp->vel[vu_pos + axr_factor] = etasp->vel[vu_pos];
      etasp->vel[vr_pos + axr_factor] = etasp->vel[vr_pos];
      Grsp->vel[vu_pos + axr_factor] = Grsp->vel[vu_pos];  
      Grsp->vel[vr_pos + axr_factor] = Grsp->vel[vr_pos];
      XS[pos+axr_factor] = XS[pos];
    } 
  }
  Grsp->vel[posu0+vu_offset] = -Grsp->vel[posu1+vu_offset];
  Grsp->vel[posu0+vr_offset] = -Grsp->vel[posu1+vr_offset];
  etasp->vel[posu0+vu_offset] = -etasp->vel[posu1+vu_offset];
  etasp->vel[posu0+vr_offset] = -etasp->vel[posu1+vr_offset];
  du = axu->r[axu_factor*(Nu+1)] - axu->r[axu_factor*Nu];
  pos = i*axr_factor + Nu*axu_factor;
  x = eta_/eta[pos-pnt*axu_factor];
  r = du/(axu->r[pos] - axu->r[pos - axu_factor*pnt]);
  Gz_TBC[i] = Gr_TBC[i] = eta_TBC[i] = pow(x,r);
}
 if (fail) { 
   eta_ = eta[jp*axu_factor+ip*axr_factor];
   fprintf(stderr,"sources: Error in eta[%u,%u]: %8.5e\n",ip,jp,eta_);
   /*   exit(1);*/
 }

for(i=1; i<= (Nr+1); i++){
  for(j=0; j<2*Nexp; j+=2){
    rhox = exsp[j]->rho[i];
    exsp[j]->vel[i] = (rhox > 0.0)?(exsp[j+1]->rho[i]/rhox):0.0;
    if (i == 1) {
      exsp[j]->vel[0] = -exsp[j]->vel[1];
    }
  }
  nisp->vel[i] = (ni[i] != 0.0)?(Gammai[i]/ni[i]):0.0;
}
nisp->vel[0] = -nisp->vel[1]; 
/*
  get excited species collision terms
*/
 for(j=0; j<2*Nexp; j+=2){
   get_heavy_wall_terms(exsp[j], &(ex_W[j]), exsp[j+1], &(ex_W[j+1]));
 }
/*
  get ion wall collision terms
*/
get_heavy_wall_terms(nisp, &ni_W, Gammaisp, &Gammai_W );
/*
  Advance variables by updating source terms
*/
MD_FCT_Richardson(set, advance_sources, dt);

for(i=1; i<= (Nr+1); i++){
  Ersp->vel[i] = Vx*SQR(axr_config->r[i])/(R*R);
}
Ersp->vel[0] = Ersp->vel[1];
/*
  Determine new time step

  get maximum Er
  Take conservative (noncollisional) time step
  restriction from (du/E*v) and (dr/v).
  Then do ion contribution.

*/
maxEr = 0.0; 
for(i=1; i<=Nr; i++){
  absEr = ABS(Er[i]);
  maxEr = max(maxEr, absEr);
}
maxEs = SQR(maxEr) + SQR(Ez);
MD_Source_dt = dtkinetic;
for(i=1; i<= Nu; i++){
  c = beta[i];
  d = alpha[i]*maxEs;
  tmpdt = (c + sqrt(c*c+2.0*d))/d;
  MD_Source_dt = min(MD_Source_dt, tmpdt);
}

a = ion_a1*maxEr + ion_a2;

tmpdt = (ion_elastic + sqrt(ion_elastic*ion_elastic + 4*a))/(2*a);

MD_Source_dt = min(MD_Source_dt, tmpdt);
/*
  print status and log data
*/
if (MD_FCT_step == half){
  fprintf(stderr,"MD_Source_dt: %8.5e total time: %8.5e ",MD_Source_dt, MD_Total_time);
  fprintf(stderr,"Ez: %8.5e I: %8.5e\n", Ez,I);
}
return;
}

